/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor;

import static org.eclipse.core.runtime.IStatus.ERROR;
import static org.eclipse.jface.dialogs.ErrorDialog.openError;
import static org.eclipse.swt.widgets.Display.getCurrent;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.DEFAULT_CONFIGURATION_NAME;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.getActiveConfigurationsTransitively;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.getActiveConstraintsTransitively;
import static org.fortiss.tooling.base.utils.ConstraintsBaseUtils.retrieveDefaultConfig;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import org.conqat.ide.commons.ui.jface.TreeContentProviderBase;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.CheckStateChangedEvent;
import org.eclipse.jface.viewers.CheckboxTreeViewer;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.ICheckStateListener;
import org.eclipse.jface.viewers.ICheckStateProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.jface.viewers.TreeViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.TreeColumn;
import org.fortiss.tooling.base.model.element.ConstraintConfiguration;
import org.fortiss.tooling.base.model.element.ElementFactory;
import org.fortiss.tooling.base.model.element.IConstraintBasedProcess;
import org.fortiss.tooling.base.ui.ToolingBaseUIActivator;
import org.fortiss.tooling.base.ui.utils.ConstraintsBaseUIUtils;
import org.fortiss.tooling.base.ui.utils.StatusUtils;
import org.fortiss.tooling.kernel.extension.IConstraint;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.service.IConstraintService;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.extension.base.EditorBase;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;
import org.fortiss.tooling.kernel.ui.service.IActionService;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService;
import org.fortiss.tooling.kernel.ui.service.IContextMenuService;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Editor for {@link IConstraintBasedProcess}s.
 * 
 * TODO(2931): should not be an editor but a widget
 * 
 * @author aravantinos
 */
public class ConstraintBasedProcessEditor<CBP extends IConstraintBasedProcess> extends
		EditorBase<CBP> implements ContextMenuContextProvider, ISelectionProvider {

	/** {@inheritDoc} */
	@Override
	public void addPartPropertyListener(IPropertyChangeListener listener) {
		super.addPartPropertyListener(listener);
	}

	/** Model context of the edited project. */
	private ITopLevelElement top;

	/** Constraint container for the edited project. */
	private IConstraintInstanceContainer cstrContainer;

	/** Tree viewer for the configurations. */
	private CheckboxTreeViewer treeViewer;

	/** Selection of the tree. */
	private ISelection selection;

	/** Adapter for refreshing content. */
	private EContentAdapter refreshAdapter;

	/** Reference to a constraint configuration. */
	private static class WithContextualConfiguration {

		/** Contextual constraint configuration. */
		ConstraintConfiguration config;

		/** Constructor. */
		public WithContextualConfiguration(ConstraintConfiguration config) {
			this.config = config;
		}
	}

	/** Constraint accompanied with a configuration providing the context. */
	private static class ConstraintWithContextualConfiguration extends WithContextualConfiguration {

		/** Constraint. */
		Class<? extends IConstraint> cstr;

		/** Constructor. */
		public ConstraintWithContextualConfiguration(Class<? extends IConstraint> cstr,
				ConstraintConfiguration config) {
			super(config);
			this.cstr = cstr;
		}
	}

	/**
	 * Configuration accompanied with another configuration providing the context.
	 * Intent is that the contained configuration is a dependency of the context one.
	 */
	private static class ConfigurationRefWithContextualConfiguration extends
			WithContextualConfiguration {

		/** Constraint. */
		ConstraintConfiguration target;

		/** Constructor. */
		public ConfigurationRefWithContextualConfiguration(ConstraintConfiguration configRef,
				ConstraintConfiguration config) {
			super(config);
			this.target = configRef;
		}
	}

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		top = IPersistencyService.getInstance().getTopLevelElementFor(editedObject);
		getSite().setSelectionProvider(this);
		retrieveDefaultConfig(editedObject);
		cstrContainer = editedObject.getConstraintInstanceContainer();
		parent.setLayout(new GridLayout(1, true));
		createTree(parent);
		MenuManager menuManager = IContextMenuService.getInstance().createDefaultContextMenu(this);
		Menu menu = menuManager.createContextMenu(treeViewer.getTree());
		treeViewer.getTree().setMenu(menu);
		createCurrentObjectivePart(parent);
		refreshAdapter = new EContentAdapter() {
			@Override
			public void notifyChanged(Notification notification) {
				treeViewer.refresh();
			}
		};
		editedObject.eAdapters().add(refreshAdapter);
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		EcoreUtils.disableNotificationsOf(editedObject);
		editedObject.eAdapters().remove(refreshAdapter);
		EcoreUtils.enableNotificationsOf(editedObject);
		super.dispose();
	}

	/** Create current the controls "Current objective:". */
	private void createCurrentObjectivePart(Composite parent) {
		Composite currentObjComposite = new Composite(parent, SWT.None);
		currentObjComposite.setLayout(new GridLayout(2, false));
	}

	/** Fake child used to add new elements. */
	private static class FakeChild {
		// Nothing to implement: just a marker class
	}

	/** Content provider for the configuration tree. */
	private class ConfigTreeContentProvider extends TreeContentProviderBase {

		/** {@inheritDoc} */
		@Override
		public Object[] getChildren(Object parentElement) {
			if(parentElement instanceof IConstraintBasedProcess) {
				IConstraintBasedProcess cbdp = (IConstraintBasedProcess)parentElement;
				ArrayList<Object> configs = new ArrayList<>();
				configs.add(new FakeChild());
				configs.addAll(cbdp.getConfigurations());
				return configs.toArray();
			} else if(parentElement instanceof ConstraintConfiguration) {
				List<Object> types = new ArrayList<Object>();
				Set<Class<? extends IConstraint>> availableCstrs =
						IConstraintUIService.getInstance().getAllConstraints();
				availableCstrs.removeAll(IConstraintUIService.getInstance()
						.getAlwaysActivatedConstraints());

				if(editedObject.getConfigurations().size() > 1) {
					List<ConfigurationRefWithContextualConfiguration> refs =
							editedObject
									.getConfigurations()
									.stream()
									.map(t -> new ConfigurationRefWithContextualConfiguration(t,
											(ConstraintConfiguration)parentElement))
									.collect(Collectors.toList());
					types.add(refs);
				}

				Set<Class<? extends IConstraint>> availableConstraints =
						IConstraintUIService.getInstance().getAllConstraints();
				IConstraintUIService cuis = IConstraintUIService.getInstance();

				// Collecting the groups and putting the constraints in the groups in the array list
				List<String> groupName = new ArrayList<String>();
				for(Class<? extends IConstraint> c : availableConstraints) {
					String constGroupName = IConstraintService.getInstance().getGroupName(c);
					if(!groupName.contains(constGroupName)) {
						groupName.add(constGroupName);
						types.add(constGroupName);
						HashMap<String, String> constraintInfo = new HashMap<String, String>();
						for(Class<? extends IConstraint> ctr : availableConstraints) {
							String ctrGroupName =
									IConstraintService.getInstance().getGroupName(ctr);
							String desc = cuis.getDescription(ctr);
							constraintMap.put(ctr.getName(), desc);
							if(ctrGroupName.equalsIgnoreCase(constGroupName)) {
								constraintInfo.put(ctr.getName(), cuis.getDescription(ctr));
							}
						}
						groupNameAndConstraints.put(constGroupName, constraintInfo);
					}
				}
				return types.toArray();
			} else if(parentElement instanceof List<?>) {
				return ((List<?>)parentElement).toArray();
			}
			// Add the constraints to the tree based on the group name.
			else if(parentElement instanceof String) {
				// Loop over all the names of the groups
				for(String constraintGroupName : groupNameAndConstraints.keySet()) {
					if(parentElement.toString().equalsIgnoreCase(constraintGroupName.toString())) {
						// Return complete keySet of the HashMap inside the value field of the
						// HashMap groupNameAndConstraints<String, "HashMap<String",String>> that
						// contains the names of the constraints that are grouped under a particular
						// constraint group
						return groupNameAndConstraints.get(constraintGroupName).keySet().toArray();
					}
				}
			}
			return new Object[0];
		}
	}

	/** This constraint map contains the name and the description of all of the constraints */
	HashMap<String, String> constraintMap = new HashMap<String, String>();

	/** This HashMap contains the groups and the the grouped constraints (as a HashMap) */
	HashMap<String, HashMap<String, String>> groupNameAndConstraints =
			new HashMap<String, HashMap<String, String>>();

	/** Text labels for the tree of configurations. */
	private String getLabelText(Object element) {
		if(element instanceof FakeChild) {
			return "click to add a new configuration";
		} else if(element instanceof ConstraintConfiguration) {
			return ((ConstraintConfiguration)element).getName();
		} else if(element instanceof ConstraintWithContextualConfiguration) {
			IConstraintUIService cuis = IConstraintUIService.getInstance();
			return cuis.getDescription(((ConstraintWithContextualConfiguration)element).cstr);
		} else if(element instanceof ConfigurationRefWithContextualConfiguration) {
			return ((ConfigurationRefWithContextualConfiguration)element).target.getName();
		} else if(element instanceof List<?>) {
			return "Required configurations";
		} else if(element instanceof String) {
			if(groupNameAndConstraints.containsKey(element)) {
				for(String constraintGroupName : groupNameAndConstraints.keySet()) {
					if(element.toString().equalsIgnoreCase(constraintGroupName.toString())) {
						return constraintGroupName;
					}
				}
			} else if(constraintMap.containsKey(element)) {
				return constraintMap.get(element);
			}
		}
		return null;
	}

	/** Check state provider for the tree of configurations. */
	private class ConfigTreeCheckStateProvider implements ICheckStateProvider {

		/** {@inheritDoc} */
		@Override
		public boolean isGrayed(Object element) {
			return !(element instanceof ConstraintWithContextualConfiguration ||
					element instanceof ConfigurationRefWithContextualConfiguration || constraintMap
						.containsKey(element));
		}

		/** {@inheritDoc} */
		@Override
		public boolean isChecked(Object element) {
			if(element instanceof ConstraintWithContextualConfiguration) {
				ConstraintWithContextualConfiguration cstrInContext =
						(ConstraintWithContextualConfiguration)element;
				IConstraintService cs = IConstraintService.getInstance();
				String name = cs.getName(cstrInContext.cstr);
				return cstrInContext.config.getActiveConstraints().contains(name);
			} else if(element instanceof ConfigurationRefWithContextualConfiguration) {
				ConfigurationRefWithContextualConfiguration configRef =
						(ConfigurationRefWithContextualConfiguration)element;
				return configRef.config.getIncludedConfigurations().contains(configRef.target);
			} else if(constraintMap.containsKey(element)) {
				return editedObject.getCurrentObjective().getActiveConstraints().contains(element);
			}
			return true;
		}
	}

	/** Check state listener for the tree of configurations. */
	private class ConfigTreeCheckStateListener implements ICheckStateListener {
		/** {@inheritDoc} */
		@Override
		public void checkStateChanged(CheckStateChangedEvent event) {
			// Get object element from the state changed
			Object element = event.getElement();
			IConstraintService cs = IConstraintService.getInstance();
			IConstraintUIService cuis = IConstraintUIService.getInstance();
			ConstraintConfiguration currentObj = editedObject.getCurrentObjective();
			if(element instanceof ConstraintWithContextualConfiguration) {
				ConstraintWithContextualConfiguration cstrInContext =
						(ConstraintWithContextualConfiguration)element;
				String name = cs.getName(cstrInContext.cstr);
				top.runAsCommand(new Runnable() {

					@Override
					public void run() {
						if(event.getChecked()) {
							cstrInContext.config.getActiveConstraints().add(name);
							List<ConstraintConfiguration> activeConfigs =
									getActiveConfigurationsTransitively(currentObj, null);
							if(activeConfigs.contains(cstrInContext.config)) {
								cuis.activate(cstrInContext.cstr, cstrContainer);
							}
						} else {
							cstrInContext.config.getActiveConstraints().remove(name);
							List<Class<? extends IConstraint>> stillActiveCstrs =
									getActiveConstraintsTransitively(currentObj,
											cstrInContext.config);
							if(!stillActiveCstrs.contains(cstrInContext.cstr)) {
								cuis.deactivate(cstrInContext.cstr, cstrContainer);
							}
						}
					}
				});
			} else if(element instanceof ConfigurationRefWithContextualConfiguration) {
				ConfigurationRefWithContextualConfiguration configRef =
						(ConfigurationRefWithContextualConfiguration)element;
				top.runAsCommand(() -> {
					if(event.getChecked()) {
						configRef.config.getIncludedConfigurations().add(configRef.target);
						ConstraintsBaseUIUtils.activateConfiguration(configRef.target,
								cstrContainer);
					} else {
						configRef.config.getIncludedConfigurations().remove(configRef.target);
						ConstraintsBaseUIUtils.deactivateConfiguration(configRef.target,
								cstrContainer, configRef.config);
					}
				});
			}
			// Check if the constraint belongs to the constraints HashMap
			else if(constraintMap.containsKey(element)) {

				top.runAsCommand(new Runnable() {
					@Override
					public void run() {
						// Get the constraintName as a full class path (e.g.,
						// a.b.c.constraintName.Java)
						String constraintName = element.toString();
						if(event.getChecked()) {
							// If the constraint is checked than add it to the active
							// constraints list
							editedObject.getCurrentObjective().getActiveConstraints()
									.add(constraintName);
							// Collect all the active constraints in the list of constraints
							List<ConstraintConfiguration> activeConfigs =
									getActiveConfigurationsTransitively(currentObj, null);
							// If the selected constraint is in the active constraints list
							EList<String> activeConstraints =
									activeConfigs.get(0).getActiveConstraints();
							if(activeConstraints.contains(cs.getConstraintByName(constraintName)
									.getCanonicalName())) {
								// Activate the constraint
								cuis.activate(cs.getConstraintByName(constraintName), cstrContainer);
							}
						} else {
							// If the constraint is unchecked from the tree remove the
							// constraint from
							// the active constraints list
							editedObject.getCurrentObjective().getActiveConstraints()
									.remove(constraintName);
							List<ConstraintConfiguration> stillActiveConfigs =
									getActiveConfigurationsTransitively(currentObj, null);
							// If the constraint is still not active
							EList<String> stillActiveConstraints =
									stillActiveConfigs.get(0).getActiveConstraints();
							if(!stillActiveConstraints.contains(cs.getConstraintByName(
									constraintName).getCanonicalName())) {
								Class<? extends IConstraint> constrainttoDeActivate =
										cs.getConstraintByName(constraintName);
								// Deactivate the constraint
								cuis.deactivate(constrainttoDeActivate, cstrContainer);
							}
						}
					}
				});
			}
		}
	}

	/** Tree for configurations. */
	private class ConfigTree extends CheckboxTreeViewer implements ISelectionChangedListener {

		/** {@inheritDoc} */
		@Override
		public void addSelectionChangedListener(ISelectionChangedListener listener) {
			super.addSelectionChangedListener(listener);
		}

		/** Constructor. */
		public ConfigTree(Composite parent) {
			super(parent);
			getTree().setLayoutData(new GridData(800, 600));
			setContentProvider(new ConfigTreeContentProvider());
			setCheckStateProvider(new ConfigTreeCheckStateProvider());
			addCheckStateListener(new ConfigTreeCheckStateListener());
			addSelectionChangedListener(this);
		}

		/** {@inheritDoc} */
		@Override
		public void selectionChanged(SelectionChangedEvent event) {
			selection = event.getSelection();
			IActionService.getInstance().refresh();
		}
	}

	/** Creates the tree of configurations. */
	private void createTree(Composite parent) {
		treeViewer = new ConfigTree(parent);
		TreeColumn column = new TreeColumn(treeViewer.getTree(), SWT.LEFT);
		column.setText("Constraint configurations");
		column.setWidth(1300);
		TreeViewerColumn onlyViewerColumn = new TreeViewerColumn(treeViewer, column);
		onlyViewerColumn.setLabelProvider(new ColumnLabelProvider() {

			/** {@inheritDoc} */
			@Override
			public String getText(Object element) {
				return getLabelText(element);
			}

			@Override
			public Color getForeground(Object element) {
				return getConstraintForeGround(element, colorCtr);
			}

			@Override
			public Color getBackground(Object element) {
				// Collect all the active constraints from the currentObject (i.e., the current
				// active configuration)
				List<String> activeConstraints =
						editedObject.getCurrentObjective().getActiveConstraints();
				if(constraintMap.containsKey(element)) {
					// Ask for the constraint color based on status
					return getConstraintBackgroundColor(element, activeConstraints);
				}
				return null;
			}
		});

		onlyViewerColumn.setEditingSupport(new EditingSupport(treeViewer) {
			@Override
			protected void setValue(Object element, Object value) {
				if(element instanceof FakeChild) {
					if(!alreadyExists((String)value)) {
						ConstraintConfiguration newconfig =
								ElementFactory.eINSTANCE.createConstraintConfiguration();
						newconfig.setName((String)value);
						top.runAsCommand(() -> IElementCompositorService.getInstance().compose(
								editedObject, newconfig, null));
					}
				}
				if(element instanceof ConstraintConfiguration) {
					ConstraintConfiguration config = (ConstraintConfiguration)element;
					if(config.getName().equals(DEFAULT_CONFIGURATION_NAME) ||
							config.getName().equals(value)) {
						return;
					}
					if(!alreadyExists((String)value)) {
						top.runAsCommand(() -> ((ConstraintConfiguration)element)
								.setName((String)value));
					}
				}
			}

			@Override
			protected Object getValue(Object element) {
				return getLabelText(element);
			}

			@Override
			protected CellEditor getCellEditor(Object element) {
				return new TextCellEditor((Composite)getViewer().getControl());
			}

			@Override
			protected boolean canEdit(Object element) {
				return element instanceof ConstraintConfiguration || element instanceof FakeChild ||
						constraintMap.containsKey(element);
			}
		});
		treeViewer.setComparator(new ConstraintComparator());
		treeViewer.setInput(editedObject);
	}

	/**
	 * @param element
	 *            is the selected constraint
	 * @param activeConstraints
	 *            active constraints in the current objective
	 * @return return the color (i.e., red (Failed), green (Successful) or white(not active))
	 */
	public Color getConstraintBackgroundColor(Object element, List<String> activeConstraints) {
		// Get the complete path of the constraint class (e.g., a.b.c.constraint.java)
		String constraintToFind = element.toString();
		// If the constraint is in the active constraint list and the status is true
		// than make the background color green
		if(activeConstraints.contains(constraintToFind) &&
				StatusUtils.areAllInstancesSuccessful(constraintToFind, editedObject
						.getConstraintInstanceContainer(), editedObject.getCurrentObjective()
						.getActiveConstraints()) == true) {
			return getCurrent().getSystemColor(SWT.COLOR_GREEN);
		}
		// If the constraint doesn't belong to the active list of constrains than let
		// the color remain white
		else if(activeConstraints.contains(constraintToFind) == false) {
			return getCurrent().getSystemColor(SWT.COLOR_WHITE);
		}
		// If the constraint is active and the status is false than make the background
		// color red
		if(activeConstraints.contains(constraintToFind) &&
				StatusUtils.areAllInstancesSuccessful(constraintToFind, editedObject
						.getConstraintInstanceContainer(), editedObject.getCurrentObjective()
						.getActiveConstraints()) == false) {
			return getCurrent().getSystemColor(SWT.COLOR_RED);
		}
		return null;
	}

	/** Color Array for at least seven groups. */
	int[] colorArray = {SWT.COLOR_DARK_GREEN, SWT.COLOR_DARK_BLUE, SWT.COLOR_DARK_YELLOW,
			SWT.COLOR_BLACK, SWT.COLOR_DARK_RED, SWT.COLOR_GREEN, SWT.COLOR_CYAN};

	/** Counter for the colorArray */
	int colorCtr = 0;

	/**
	 * @param element
	 *            the groupName for which we need to set the background.
	 * @param colorIndex
	 *            index used for the colors in the colorArray.
	 * @return color required for a particular group.
	 */
	public Color getConstraintForeGround(Object element, int colorIndex) {
		if(groupNameAndConstraints.keySet().contains(element)) {
			if(colorIndex != colorCtr + 1) {
				if(groupNameAndConstraints.keySet().size() == colorCtr + 1) {
					colorCtr = 0;
				} else {
					colorCtr++;
				}
				return getCurrent().getSystemColor(colorArray[colorIndex]);
			}
		}
		return null;
	}

	/**
	 * Checks if the name already exists among current configurations, if so displays an error
	 * message and return true, otherwise returns false.
	 */
	private boolean alreadyExists(String name) {
		List<String> configNames =
				editedObject.getConfigurations().stream().map(c -> c.getName())
						.collect(Collectors.toList());
		if(configNames.contains(name)) {
			Status status =
					new Status(ERROR, ToolingBaseUIActivator.PLUGIN_ID, 0,
							"A configuration with this name already exists.", null);
			openError(getCurrent().getActiveShell(), "Error", null, status);
			return true;
		}
		return false;
	}

	/** Comparator for the contents of the tree. */
	private class ConstraintComparator extends ViewerComparator {

		/** {@inheritDoc} */
		@Override
		public int compare(Viewer viewer, Object e1, Object e2) {
			if(e1 instanceof FakeChild) {
				return -1;
			} else if(e2 instanceof FakeChild) {
				return 1;
			} else if(e1 instanceof ConstraintConfiguration &&
					e2 instanceof ConstraintConfiguration) {
				return ((ConstraintConfiguration)e1).getName().compareToIgnoreCase(
						((ConstraintConfiguration)e2).getName());
			} else if(e1 instanceof ConstraintWithContextualConfiguration &&
					e2 instanceof ConstraintWithContextualConfiguration) {
				IConstraintUIService cuis = IConstraintUIService.getInstance();
				String desc1 =
						cuis.getDescription(((ConstraintWithContextualConfiguration)e1).cstr);
				String desc2 =
						cuis.getDescription(((ConstraintWithContextualConfiguration)e2).cstr);
				return desc1.compareToIgnoreCase(desc2);
			} else if(e1 instanceof ConfigurationRefWithContextualConfiguration &&
					e2 instanceof ConfigurationRefWithContextualConfiguration) {
				String n1 = ((ConfigurationRefWithContextualConfiguration)e1).target.getName();
				String n2 = ((ConfigurationRefWithContextualConfiguration)e2).target.getName();
				return n1.compareToIgnoreCase(n2);
			}
			return super.compare(viewer, e1, e2);
		}
	}

	/** {@inheritDoc} */
	@Override
	public EObject getSelectedModelElement() {
		if(getSelectedModelElementList().isEmpty()) {
			return null;
		}
		// Returns the first element, which is ensured to exist by the previous statement.
		Object selected = getSelectedModelElementList().get(0);
		return selected instanceof EObject ? (EObject)selected : null;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public List<EObject> getSelectedModelElementList() {
		boolean isStructured = selection instanceof StructuredSelection;
		return isStructured ? ((StructuredSelection)selection).toList() : new ArrayList<>();
	}

	/** {@inheritDoc} */
	@Override
	public void addSelectionChangedListener(ISelectionChangedListener listener) {
		// Not supported
	}

	/** {@inheritDoc} */
	@Override
	public ISelection getSelection() {
		return selection == null ? new StructuredSelection() : selection;
	}

	/** {@inheritDoc} */
	@Override
	public void removeSelectionChangedListener(ISelectionChangedListener listener) {
		// Not supported
	}

	/** {@inheritDoc} */
	@Override
	public void setSelection(ISelection selection) {
		// Not supported
	}
}
