/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor;

import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTOR_OFFSET_ORIENTATION;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getConnectorAngle;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.setConnectorAngle;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.moveNode;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setStickyConnectorLayoutData;
import static org.fortiss.tooling.base.utils.OffsetOrientationUtils.getOffsetOrientation;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.draw2d.FigureCanvas;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.gef.GraphicalViewer;
import org.eclipse.gef.KeyHandler;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.fortiss.tooling.base.model.layout.Angle;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.LayoutFactory;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.ui.editpart.PositionedEditPartBase;
import org.fortiss.tooling.base.ui.editpart.StickyConnectorEditPartBase;
import org.fortiss.tooling.base.utils.LayoutDataUtils;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * A {@link KeyHandler} for Diagram editors.
 * 
 * Provides arrow keys handling:
 * - if an element is selected, then using the arrows will move this element
 * - if no element is selected, then using the arrows will scroll the diagram
 * In both cases, pressing the Shift key at the same time allows to move/scroll faster.
 * 
 * @author aravantinos
 */
public class DiagramKeyHandler extends KeyHandler {

	/** The embedded viewer. */
	private GraphicalViewer viewer;

	/** Constructor. */
	public DiagramKeyHandler(GraphicalViewer viewer) {
		this.viewer = viewer;
	}

	/** {@inheritDoc} */
	@Override
	public boolean keyPressed(KeyEvent event) {
		if(event.keyCode != SWT.ARROW_DOWN && event.keyCode != SWT.ARROW_UP &&
				event.keyCode != SWT.ARROW_LEFT && event.keyCode != SWT.ARROW_RIGHT) {
			return false;
		}
		if(scrollViewerIfNeeded(event)) {
			return true;
		}
		return moveElementIfNeeded(event);
	}

	/** Scrolls the viewer according to the event if no editpart is selected. */
	private boolean scrollViewerIfNeeded(KeyEvent event) {
		if(!(viewer.getSelectedEditParts().isEmpty() && viewer.getControl() instanceof FigureCanvas))
			return false;
		FigureCanvas figCanvas = (FigureCanvas)viewer.getControl();
		Point loc = figCanvas.getViewport().getViewLocation();
		Rectangle area = figCanvas.getViewport().getClientArea(Rectangle.SINGLETON).scale(.1);
		Point scrollCoeff = computeScrollCoeff(event);
		figCanvas.scrollToY(loc.y + scrollCoeff.y * area.height);
		figCanvas.scrollToX(loc.x + scrollCoeff.x * area.width);
		return true;
	}

	/** Scrolls the model of the selected editpart according to the event. */
	@SuppressWarnings("unchecked")
	private boolean moveElementIfNeeded(KeyEvent event) {
		List<PositionedEditPartBase<?>> positionedEditParts = new ArrayList<>();
		positionedEditParts.addAll((Collection<? extends PositionedEditPartBase<?>>)pickInstanceOf(
				PositionedEditPartBase.class, viewer.getSelectedEditParts()));

		List<ILayoutedModelElement> models = new ArrayList<ILayoutedModelElement>();
		for(PositionedEditPartBase<?> ep : positionedEditParts) {
			models.add(ep.getModel());
		}

		// Filter models which are contained by other selected models: as a heuristic, in such cases
		// the user most generally only wants to move the container. The contained element will then
		// move along.
		List<PositionedEditPartBase<?>> toRemoves = new ArrayList<>();
		for(PositionedEditPartBase<?> ep : positionedEditParts) {
			if(models.contains(ep.getModel().eContainer())) {
				toRemoves.add(ep);
			}
			if(ep instanceof StickyConnectorEditPartBase<?>) {
				OffsetOrientation oo =
						getOffsetOrientation(ep.getModel(), CONNECTOR_OFFSET_ORIENTATION);
				if(oo != null) {
					if((oo.getOrientation() == EOrientation.EAST || oo.getOrientation() == EOrientation.WEST) &&
							(event.keyCode == SWT.ARROW_DOWN || event.keyCode == SWT.ARROW_UP)) {
						continue;
					}
					if((oo.getOrientation() == EOrientation.NORTH || oo.getOrientation() == EOrientation.SOUTH) &&
							(event.keyCode == SWT.ARROW_LEFT || event.keyCode == SWT.ARROW_RIGHT)) {
						continue;
					}
				}
				if(getConnectorAngle(ep.getModel()) != null) {
					continue;
				}
				toRemoves.add(ep);
			}
		}
		positionedEditParts.removeAll(toRemoves);

		if(positionedEditParts.isEmpty()) {
			return false;
		}

		// get(0) because we know the list is not empty and because we assume there is only one top
		// level element for the given editor.
		ITopLevelElement topLevel =
				IPersistencyService.getInstance().getTopLevelElementFor(models.get(0));
		final Point scrollCoeff = computeScrollCoeff(event);
		final double angleCoeff = computeScrollAngle(event);
		scrollCoeff.scale(DEFAULT_GRID_SIZE);
		topLevel.runAsCommand(new Runnable() {
			@Override
			public void run() {
				for(PositionedEditPartBase<?> ep : positionedEditParts) {
					if(ep instanceof StickyConnectorEditPartBase<?>) {
						OffsetOrientation oo =
								getOffsetOrientation(ep.getModel(), CONNECTOR_OFFSET_ORIENTATION);
						if(oo != null) {
							Dimension containerDim = LayoutFactory.eINSTANCE.createDimension();
							if(ep.getModel().eContainer() instanceof ILayoutedModelElement) {
								ILayoutedModelElement layoutedContainer =
										(ILayoutedModelElement)ep.getModel().eContainer();
								containerDim = LayoutDataUtils.getNodeSize(layoutedContainer);
							}
							if((oo.getOrientation() == EOrientation.EAST || oo.getOrientation() == EOrientation.WEST) &&
									(event.keyCode == SWT.ARROW_DOWN || event.keyCode == SWT.ARROW_UP)) {
								int newY = oo.getOffset() + scrollCoeff.y;
								moveVertical(newY, containerDim, ep.getModel(), oo.getOrientation());
							} else if((oo.getOrientation() == EOrientation.NORTH || oo
									.getOrientation() == EOrientation.SOUTH) &&
									(event.keyCode == SWT.ARROW_LEFT || event.keyCode == SWT.ARROW_RIGHT)) {
								int newX = oo.getOffset() + scrollCoeff.x;
								moveHorizontal(newX, containerDim, ep.getModel(),
										oo.getOrientation());
							}
						} else {
							Angle angle = getConnectorAngle(ep.getModel());
							if(angle != null) {
								setConnectorAngle(ep.getModel(), angle.getAngle() + angleCoeff);
							}
						}
					} else {
						moveNode(ep.getModel(), scrollCoeff.x, scrollCoeff.y);
					}
				}
			}
		});
		return true;
	}

	/** Moves the given layouted element vertically by the distance provided by offset. */
	private void moveVertical(int offset, Dimension containerDim, ILayoutedModelElement layouted,
			EOrientation orientation) {
		int newOffset =
				orientation == EOrientation.WEST ? 0 : containerDim.getWidth() - 2 *
						DEFAULT_GRID_SIZE;
		if(offset <= 0) {
			setStickyConnectorLayoutData(layouted, EOrientation.NORTH, newOffset);
		} else if(offset >= containerDim.getHeight() - 2 * DEFAULT_GRID_SIZE) {
			setStickyConnectorLayoutData(layouted, EOrientation.SOUTH, newOffset);
		} else {
			setStickyConnectorLayoutData(layouted, orientation, offset);
		}
	}

	/** Moves the given layouted element horizontally by the distance provided by offset. */
	private void moveHorizontal(int offset, Dimension containerDim, ILayoutedModelElement layouted,
			EOrientation orientation) {
		int newOffset =
				orientation == EOrientation.NORTH ? 0 : containerDim.getHeight() - 2 *
						DEFAULT_GRID_SIZE;
		if(offset <= 0) {
			setStickyConnectorLayoutData(layouted, EOrientation.WEST, newOffset);
		} else if(offset >= containerDim.getWidth() - 2 * DEFAULT_GRID_SIZE) {
			setStickyConnectorLayoutData(layouted, EOrientation.EAST, newOffset);
		} else {
			setStickyConnectorLayoutData(layouted, orientation, offset);
		}
	}

	/** Returns a coefficient of 1 if the SHIFT key is not pressed, 50 otherwise. */
	private Point computeScrollCoeff(KeyEvent event) {
		int coeff = (event.stateMask & SWT.SHIFT) == 0 ? 1 : 10;
		if(event.keyCode == SWT.ARROW_DOWN)
			return new Point(0, coeff);
		if(event.keyCode == SWT.ARROW_UP)
			return new Point(0, -coeff);
		coeff *= (viewer.getControl().getStyle() & SWT.MIRRORED) != 0 ? -1 : 1;
		if(event.keyCode == SWT.ARROW_LEFT)
			return new Point(-coeff, 0);
		if(event.keyCode == SWT.ARROW_RIGHT)
			return new Point(coeff, 0);
		return null;
	}

	/** Returns a coefficient of 1 if the SHIFT key is not pressed, 50 otherwise. */
	private double computeScrollAngle(KeyEvent event) {
		double angle = 0.2;
		if(event.keyCode == SWT.ARROW_DOWN)
			return angle;
		if(event.keyCode == SWT.ARROW_UP)
			return -angle;
		angle *= (viewer.getControl().getStyle() & SWT.MIRRORED) != 0 ? -1 : 1;
		if(event.keyCode == SWT.ARROW_LEFT)
			return -angle;
		if(event.keyCode == SWT.ARROW_RIGHT)
			return angle;
		return 0.;
	}
}
