/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editpart.policy;

import static org.fortiss.tooling.base.ui.utils.DragAndDropBaseUtils.createConnectionDragContext;

import org.eclipse.draw2d.geometry.Point;
import org.eclipse.gef.ConnectionEditPart;
import org.eclipse.gef.GraphicalEditPart;
import org.eclipse.gef.NodeEditPart;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.commands.UnexecutableCommand;
import org.eclipse.gef.editpolicies.GraphicalNodeEditPolicy;
import org.eclipse.gef.requests.CreateConnectionRequest;
import org.eclipse.gef.requests.ReconnectRequest;
import org.fortiss.tooling.base.dnd.ConnectionDragContext;
import org.fortiss.tooling.base.model.base.ConnectionSegmentBase;
import org.fortiss.tooling.base.ui.editpart.command.ConnectCommand;
import org.fortiss.tooling.base.ui.editpart.command.ReconnectCommand;

/**
 * This is the edit policy to be installed on edit parts that allow creation of
 * direct connections. This creation may also include creating connector
 * sub-nodes.
 * <p>
 * You usually install a customized drag tracker with an edit part with this policy. Here is an
 * example using CTRL key to initiate a connection creation.
 * 
 * <pre>
 * 
 * &#064;Override
 * public DragTracker getDragTracker(Request request) {
 * 	if(request instanceof SelectionRequest) {
 * 		SelectionRequest sr = (SelectionRequest)request;
 * 		if(sr.getLastButtonPressed() == 1 &amp;&amp; sr.isControlKeyPressed()) {
 * 			return new ConnectionDragCreationTool();
 * 		} else if(sr.getLastButtonPressed() == 1) {
 * 			return new DragEditPartsTracker(this);
 * 		}
 * 	}
 * 	return super.getDragTracker(request);
 * }
 * 
 * </pre>
 * <p>
 * Furthermore, to obtain correct drag feedback your edit part must implement {@link NodeEditPart}
 * and provide suitable anchor positions. You usually use the location information provided with the
 * {@link ConnectCommand} for this.
 * <p>
 * An implementation might look like this:
 * 
 * <pre>
 * public ConnectionAnchor getSourceConnectionAnchor(Request request) {
 * 	if(request instanceof CreateConnectionRequest) {
 * 		return new XYAnchor(((CreateConnectionRequest)request).getLocation());
 * 	}
 * 	return null;
 * }
 * 
 * public ConnectionAnchor getTargetConnectionAnchor(Request request) {
 * 	if(request instanceof CreateConnectionRequest) {
 * 		return new XYAnchor(((CreateConnectionRequest)request).getLocation());
 * 	}
 * 	return null;
 * }
 * </pre>
 * 
 * @author hoelzl
 */
public class DirectConnectionAndReconnectionEditPolicy extends GraphicalNodeEditPolicy {

	/** {@inheritDoc} */
	@Override
	protected Command getConnectionCompleteCommand(CreateConnectionRequest request) {
		if(request.getStartCommand() instanceof ConnectCommand) {
			Point location = new Point(request.getLocation());
			((GraphicalEditPart)getHost()).getFigure().translateToRelative(location);
			ConnectCommand cmd = (ConnectCommand)request.getStartCommand();
			cmd.setTarget(getHost(), location);
			return cmd;
		}
		return UnexecutableCommand.INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	protected Command getConnectionCreateCommand(CreateConnectionRequest request) {
		// Here we need to create a new point, otherwise the translation on the next line will
		// *modify* the request which brings problems later on.
		Point location = new Point(request.getLocation());
		((GraphicalEditPart)getHost()).getFigure().translateToRelative(location);
		ConnectionDragContext context =
				createConnectionDragContext(getHost(), getHost(), location, null);
		ConnectCommand cmd = new ConnectCommand(context);
		request.setStartCommand(cmd);
		return cmd;
	}

	/** {@inheritDoc} */
	@Override
	protected Command getReconnectSourceCommand(ReconnectRequest request) {
		if(request.getConnectionEditPart().getModel() instanceof ConnectionSegmentBase) {
			ConnectionEditPart editPart = request.getConnectionEditPart();
			ConnectionDragContext context =
					createConnectionDragContext(getHost(), editPart.getTarget(),
							request.getLocation(), null);
			ConnectionSegmentBase connection = (ConnectionSegmentBase)editPart.getModel();
			return new ReconnectCommand(context, connection);
		}
		return UnexecutableCommand.INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	protected Command getReconnectTargetCommand(ReconnectRequest request) {
		if(request.getConnectionEditPart().getModel() instanceof ConnectionSegmentBase) {
			ConnectionEditPart editPart = request.getConnectionEditPart();
			ConnectionDragContext context =
					createConnectionDragContext(editPart.getSource(), getHost(), null,
							request.getLocation());
			ConnectionSegmentBase connection =
					(ConnectionSegmentBase)request.getConnectionEditPart().getModel();
			return new ReconnectCommand(context, connection);
		}
		return UnexecutableCommand.INSTANCE;
	}
}
