/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import org.eclipse.draw2d.geometry.Point;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.RootEditPart;
import org.fortiss.tooling.base.dnd.ConnectionDragContext;
import org.fortiss.tooling.base.dnd.ConnectionDropContext;
import org.fortiss.tooling.base.ui.editpart.DiagramEditPartBase;
import org.fortiss.tooling.base.ui.editpart.FreeConnectorEditPartBase;
import org.fortiss.tooling.base.utils.LayoutModelElementFactory;
import org.fortiss.tooling.kernel.extension.data.IConnectionCompositionContext;

/**
 * Utility class to support drag and drop operations.
 * 
 * @author diewald
 */
public class DragAndDropBaseUtils {

	/**
	 * Utility method to create a {@link ConnectionDragContext} between a source and a target
	 * {@link EditPart}.
	 * 
	 * @param sourceEditPart
	 *            The source's {@link EditPart}.
	 * @param targetEditPart
	 *            The target's {@link EditPart}.
	 * @param sourcePosition
	 *            Position of the source {@link EditPart}.
	 * @param targetPosition
	 *            Position of the target {@link EditPart}.
	 * @return {@link ConnectionDragContext} which is free of UI-dependent parts.
	 */
	public static ConnectionDragContext createConnectionDragContext(EditPart sourceEditPart,
			EditPart targetEditPart, Point sourcePosition, Point targetPosition) {
		Object sourceModelObj = sourceEditPart.getModel();
		Object targetModelObj = targetEditPart.getModel();
		if((sourceModelObj instanceof EObject && targetModelObj instanceof EObject)) {
			EObject sourceParent = ((EObject)sourceModelObj).eContainer();
			EObject targetParent = ((EObject)targetModelObj).eContainer();
			if(sourceParent != null && targetParent != null) {
				org.fortiss.tooling.base.model.layout.Point sourceModelPoint =
						LayoutModelElementFactory.createPoint(sourcePosition.x, sourcePosition.y,
								sourcePosition.toString() + "_sourcePos");
				ConnectionDragContext context =
						new ConnectionDragContext((EObject)sourceModelObj, sourceParent,
								sourceModelPoint, isConsideredAsRootEditPart(sourceEditPart),
								sourceEditPart instanceof FreeConnectorEditPartBase);
				org.fortiss.tooling.base.model.layout.Point targetModelPoint =
						LayoutModelElementFactory.createPoint(sourcePosition.x, sourcePosition.y,
								sourcePosition.toString() + "_targetPos");
				context.setTarget((EObject)targetModelObj, targetParent, targetModelPoint,
						targetEditPart instanceof FreeConnectorEditPartBase,
						isConsideredAsRootEditPart(targetEditPart));
				return context;
			}
		}
		return null;
	}

	/**
	 * @param targetEditPart
	 * @return <code>true</code> if <code>targetEditPart</code> or its parent is a
	 *         {@link RootEditPart}.
	 */
	private static boolean isConsideredAsRootEditPart(EditPart targetEditPart) {
		return targetEditPart instanceof RootEditPart ||
				targetEditPart.getParent() instanceof RootEditPart;
	}

	/**
	 * Utility method to create a {@link ConnectionDropContext} expressing the drop action of the
	 * model element associated with the {#link editPart}.
	 * 
	 * @param editPart
	 *            The {@link EditPart} which is dropped.
	 * @param position
	 *            {@link Point} at which the {@link EditPart} is dropped.
	 * @return {@link ConnectionDropContext} describing the drop action and which is free of
	 *         UI-dependent parts.
	 */
	public static ConnectionDropContext createConnectionDropContext(EditPart editPart,
			Point position) {
		Object modelObj = editPart.getModel();
		if(modelObj instanceof EObject) {
			org.fortiss.tooling.base.model.layout.Point modelPoint =
					LayoutModelElementFactory.createPoint(position.x, position.y,
							position.toString() + "_sourcePos");
			double zoom = ZoomUIUtils.determineZoom(editPart.getRoot());
			return new ConnectionDropContext((EObject)modelObj, modelPoint,
					editPart instanceof DiagramEditPartBase, zoom);
		}
		return null;
	}

	/**
	 * Utility method to set the target elements of a {@link ConnectionDragContext} which is
	 * required when the target of {@link IConnectionCompositionContext} is not known at the time of
	 * context creation.
	 * 
	 * @param context
	 *            {@link ConnectionDragContext} whose target shall be set.
	 * @param editPart
	 *            The target {@link EditPart}.
	 * @param position
	 *            The position of the target.
	 * @return The given and modified {#link context}.
	 */
	public static ConnectionDragContext setConnectionDragContextTarget(
			ConnectionDragContext context, EditPart editPart, Point position) {
		Object modelObj = editPart.getModel();
		if(modelObj instanceof EObject) {
			EObject parentObj = ((EObject)modelObj).eContainer();
			org.fortiss.tooling.base.model.layout.Point modelPoint =
					LayoutModelElementFactory.createPoint(position.x, position.y,
							position.toString() + "_targetPos");
			boolean isTargetFreeConnector = editPart instanceof FreeConnectorEditPartBase;
			context.setTarget((EObject)modelObj, parentObj, modelPoint, isTargetFreeConnector,
					isConsideredAsRootEditPart(editPart));
		}
		return context;
	}
}
