/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import org.eclipse.jface.layout.TableColumnLayout;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.ColumnWeightData;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.ICellEditorValidator;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;

/**
 * Utility methods for JFace {@link TableViewer}s.
 * 
 * @author hoelzl
 */
public class TableViewerUtils {

	/**
	 * Creates a simple column in the table viewer.
	 * 
	 * @param viewer
	 *            the table viewer
	 * @param width
	 *            the column width
	 * @param title
	 *            the column title
	 * @param tooltip
	 *            the column tool tip
	 * @param labelProvider
	 *            the column's label provider
	 * @return the configured viewer column
	 */
	public static TableViewerColumn createColumn(TableViewer viewer, int width, String title,
			String tooltip, ColumnLabelProvider labelProvider) {
		TableViewerColumn column = new TableViewerColumn(viewer, SWT.LEFT);
		column.getColumn().setWidth(width);
		column.getColumn().setText(title);
		column.getColumn().setToolTipText(tooltip);
		column.setLabelProvider(labelProvider);
		return column;
	}

	/**
	 * Creates a weighted column in the table viewer.
	 * 
	 * @param viewer
	 *            the table viewer
	 * @param widthWeight
	 *            the percentage of this column width from the total width of the table
	 * @param title
	 *            the column title
	 * @param tooltip
	 *            the column tool tip
	 * @param labelProvider
	 *            the column's label provider
	 * @return the configured viewer column
	 */
	public static TableViewerColumn createWeightedColumn(TableViewer viewer, int widthWeight,
			String title, String tooltip, ColumnLabelProvider labelProvider) {
		TableViewerColumn column = new TableViewerColumn(viewer, SWT.LEFT);
		column.getColumn().setText(title);
		column.getColumn().setToolTipText(tooltip);
		column.setLabelProvider(labelProvider);

		Composite viewerParent = viewer.getTable().getParent();
		TableColumnLayout tableColumnLayout;
		if(viewerParent.getLayout() instanceof TableColumnLayout) {
			tableColumnLayout = (TableColumnLayout)viewerParent.getLayout();
		} else {
			tableColumnLayout = new TableColumnLayout();
		}

		tableColumnLayout.setColumnData(column.getColumn(), new ColumnWeightData(widthWeight));
		viewerParent.setLayout(tableColumnLayout);

		return column;
	}

	/**
	 * Creates a weighted column in the table viewer with editing support.
	 * 
	 * @param viewer
	 *            the table viewer
	 * @param widthWeight
	 *            the percentage of this column width from the total width of the table
	 * @param title
	 *            the column title
	 * @param tooltip
	 *            the column tool tip
	 * @param labelProvider
	 *            the column's label provider
	 * @param editingSupport
	 *            the column's editing support
	 * @return the configured viewer column
	 */
	public static TableViewerColumn createWeightedColumn(TableViewer viewer, int widthWeight,
			String title, String tooltip, ColumnLabelProvider labelProvider,
			EditingSupport editingSupport) {
		TableViewerColumn column = createColumn(viewer, widthWeight, title, tooltip, labelProvider);
		column.setEditingSupport(editingSupport);
		return column;
	}

	/**
	 * Creates a simple column in the table viewer with editing support.
	 * 
	 * @param viewer
	 *            the table viewer
	 * @param width
	 *            the column width
	 * @param title
	 *            the column title
	 * @param tooltip
	 *            the column tool tip
	 * @param labelProvider
	 *            the column's label provider
	 * @param editingSupport
	 *            the column's editing support
	 * 
	 * @return the configured viewer column
	 */
	public static TableViewerColumn createColumn(TableViewer viewer, int width, String title,
			String tooltip, ColumnLabelProvider labelProvider, EditingSupport editingSupport) {
		TableViewerColumn column = createColumn(viewer, width, title, tooltip, labelProvider);
		column.setEditingSupport(editingSupport);
		return column;
	}

	/** Editing support for integers. */
	public static abstract class IntegerEditingSupport extends EditingSupport {

		/** The parent where the editing support is needed. */
		private Composite parent;

		/** Constructor. */
		public IntegerEditingSupport(ColumnViewer viewer, Composite parent) {
			super(viewer);
			this.parent = parent;
		}

		/** {@inheritDoc} */
		@Override
		protected CellEditor getCellEditor(Object element) {
			CellEditor ce = new TextCellEditor(parent);
			ce.setValidator(new ICellEditorValidator() {

				/** {@inheritDoc} */
				@Override
				public String isValid(Object value) {
					try {
						Integer.parseInt(value.toString());
					} catch(NumberFormatException nfe) {
						return "Not a valid integer";
					}
					return null;
				}
			});
			return ce;
		}

		/** {@inheritDoc} */
		@Override
		protected boolean canEdit(Object element) {
			return true;
		}
	}

	/** Editing support for numbers. */
	public static abstract class NumberEditingSupport extends EditingSupport {

		/** The parent where the editing support is needed. */
		private Composite parent;

		/** Constructor. */
		public NumberEditingSupport(ColumnViewer viewer, Composite parent) {
			super(viewer);
			this.parent = parent;
		}

		/** {@inheritDoc} */
		@Override
		protected CellEditor getCellEditor(Object element) {
			CellEditor ce = new TextCellEditor(parent);
			ce.setValidator(new ICellEditorValidator() {

				/** {@inheritDoc} */
				@Override
				public String isValid(Object value) {
					if(getNumberFromString(value.toString()) == null) {
						return "Not a valid number";
					}
					return null;
				}
			});
			return ce;
		}

		/** {@inheritDoc} */
		@Override
		protected boolean canEdit(Object element) {
			return true;
		}

		/**
		 * Parses a number from a string and returns it or null if parse error
		 * occurred.
		 */
		protected Number getNumberFromString(String numAsString) {
			try {
				Number num = Integer.parseInt(numAsString);
				return num;
			} catch(NumberFormatException nfe) {
				try {
					Number num = Double.parseDouble(numAsString);
					return num;
				} catch(NumberFormatException e) {
					e.printStackTrace();
				}
			}
			return null;
		}
	}
}
