/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation.valueprovider;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.model.element.IDerivedAnnotation;

/**
 * Base class for {@link IAnnotationValueProvider} for derived annotations (i.e., based on
 * {@link IDerivedAnnotation}).
 * 
 * @author barner
 */
public abstract class DerivedAnnotationValueProviderBase<T extends IDerivedAnnotation<?>> extends
		EStructuralFeatureValueProviderBase<T> {

	/** Key representing the derived / calculated value. */
	private String derivedKey;

	/** Flag indicating whether the derived value shall be editable by the user. */
	private boolean isEditableByUser = false;

	/** List of all supported instance keys. */
	private Set<String> keys = new LinkedHashSet<>();

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a pure derived annotation (i.e., an
	 * {@link IDerivedAnnotation} that does not contain any additional {@link EStructuralFeature}s).
	 */
	public DerivedAnnotationValueProviderBase(EClass annotatedSpecificationEClass) {
		super(annotatedSpecificationEClass);
		handlesMultipleEstructuralFeatures = false;
		derivedKey = DEFAULT_KEY;
		keys.add(DEFAULT_KEY);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a pure derived annotations (i.e., an
	 * {@link IDerivedAnnotation} that does not contain any additional {@link EStructuralFeature}s).
	 */
	public DerivedAnnotationValueProviderBase(EClass annotatedSpecificationEClass,
			boolean isEditableByUser) {
		this(annotatedSpecificationEClass);
		this.isEditableByUser = isEditableByUser;
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a derived annotations (i.e., based on an
	 * {@link IDerivedAnnotation}) that contains additional {@link EStructuralFeature}s).
	 * 
	 * @param annotatedSpecificationEClass
	 *            {@link EClass} implementing the annotation
	 * @param structuralFeatureDescriptorMap
	 *            map: instance key -> {@link EStructuralFeatureDescriptor}s
	 * @param derivedKey
	 *            key representing the derived / calculated value (must not be contained in
	 *            structuralFeatureMap).
	 */
	public DerivedAnnotationValueProviderBase(EClass annotatedSpecificationEClass,
			Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap,
			String derivedKey) {
		super(annotatedSpecificationEClass, structuralFeatureDescriptorMap);

		this.derivedKey = derivedKey;
		keys.addAll(structuralFeatureDescriptorMap.keySet());
		keys.add(derivedKey);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a derived annotations (i.e., based on an
	 * {@link IDerivedAnnotation}) that contains additional {@link EStructuralFeature}s.
	 * 
	 * @param annotatedSpecificationEClass
	 *            {@link EClass} implementing the annotation
	 * @param structuralFeatureDescriptorMap
	 *            map: instance key -> {@link EStructuralFeatureDescriptor}s
	 * @param derivedKey
	 *            Key representing the derived / calculated value (must be named after the
	 *            {@link EAttribute} which shall hold the values entered by the user iff the derived
	 *            value shall be editable).
	 * @param derivedValueIsEditable
	 *            Flag, indicating whether the derived value shall be (in principal) editable by the
	 *            user.
	 */
	public DerivedAnnotationValueProviderBase(EClass annotatedSpecificationEClass,
			Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap,
			String derivedKey, boolean derivedValueIsEditable) {
		super(annotatedSpecificationEClass, structuralFeatureDescriptorMap);

		this.derivedKey = derivedKey;
		this.isEditableByUser = derivedValueIsEditable;
		keys.addAll(structuralFeatureDescriptorMap.keySet());
		keys.add(derivedKey);
	}

	/**
	 * Predicate if the given instanceKey is the key representing the derived / calculated value
	 * provided by this {@link IAnnotationValueProvider}.
	 */
	public boolean isDerivedKey(String instanceKey) {
		return (instanceKey == DEFAULT_KEY) || (instanceKey.equals(derivedKey));
	}

	/** Predicate whether the derived value shall be editable by the user. */
	public boolean isEditableByUser() {
		return isEditableByUser;
	}

	/** {@inheritDoc} */
	@Override
	public boolean canEdit(T specification, String instanceKey) {
		// Deactivate editing if...
		if(isDerivedKey(instanceKey) && !isEditableByUser) {
			// ...editing is principally disabled or...
			return false;
		}

		boolean preferDerivedValue =
				isEditableByUser && !specification.isUserAnnotatedValuePreferred();
		boolean derivedValueAvailable = specification.getDerivedValue() != null;

		if(preferDerivedValue && derivedValueAvailable) {
			// ...the derived value is available and it should override the user annotated one.
			return false;
		}
		return super.canEdit(specification, instanceKey);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		U calculatedValue;

		if(isDerivedKey(instanceKey)) {
			calculatedValue = (U)specification.getValue();
			if(calculatedValue != null) {
				return calculatedValue;
			}
		}

		return super.getAnnotationValue(specification, instanceKey);
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception {
		if(isDerivedKey(instanceKey) && specification.getDerivedFeature() != null) {
			if(value instanceof String) {
				setAnnotationValueFromString((String)value, specification, instanceKey);
			} else {
				specification.eSet(specification.getDerivedFeature(), value);
			}
			return;
		}

		super.setAnnotationValue(value, specification, instanceKey);
	}

	/** {@inheritDoc} */
	@Override
	public List<String> getInstanceKeys(T specification) {
		return new ArrayList<>(keys);
	}
}
