/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation.valueprovider;

import static org.fortiss.tooling.base.utils.BaseMathUtils.isExpectedType;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.BiConsumer;

import org.conqat.lib.commons.collections.Pair;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.service.base.IEObjectAware;

/**
 * <p>
 * Value provider interface for {@link IAnnotatedSpecification}s that establishes the link between
 * the model (data storage) and the UI (data exchange and conversion). Concrete annotations need to
 * provide specific implementations of this interface to perform the data exchange and conversion,
 * and to provide the required factory methods.
 * </p>
 * Instance keys:
 * <ul>
 * <li>Since multiple instances can exist for a single {@link IAnnotatedSpecification} (dynamic
 * instantiation by the user (see {@link DynamicInstanceAnnotationValueProviderBase}), <b>instance
 * keys</b> are used to differentiate the different instances.</li>
 * <li>Additionally, keys are used to differentiate the different annotations that can be hosted by
 * a single {@link IAnnotatedSpecification}. A typical use case is a single
 * {@link EStructuralFeatureValueProviderBase} that supports multiple (or all)
 * {@link EStructuralFeature} provided by single annotation class.</li>
 * <li>In case the annotation only provides a single value, the {@link #DEFAULT_KEY} is used. For
 * convenience, the method interface provides wrapper methods that omits the {@code instanceKey}
 * parameter and wrap the corresponding generic method using the {@link #DEFAULT_KEY}).</li>
 * </ul>
 * </p>
 * 
 * <p>
 * The interface / some of its methods take two template parameters:
 * 
 * <ul>
 * <li><tt>T</tt>: Meta-model type of the specification, i.e., the particular subclass of
 * {@link IAnnotatedSpecification} supported by the respective implementation of this
 * {@link IAnnotationValueProvider} interface.</li>
 * <li><tt>U</tt>: Data type used in the UI to support the respective annotation. For <tt>U</tt>, a
 * reasonable {@code toString()} implementation must be provided that is used by the framework to
 * render the value of the annotation in text-based editing support UI controls.</li>
 * </ul>
 * </p>
 * 
 * @author eder, diewald, barner
 */
public interface IAnnotationValueProvider<T extends IAnnotatedSpecification> extends
		IEObjectAware<IModelElement> {

	/**
	 * Container to define an action entry (name & applicable items) that typically appears in the
	 * context menu of an annotation of the annotation view.
	 */
	public class AnnotationActionEntry {
		/** Scope defining the {@link AnnotationEntry}s to apply the corresponding action to. */
		public enum ActionScope {
			/** All {@link AnnotationEntry}s that are matching the selected filter. */
			ALL_VISIBLE_ITEMS,
			/** Apply the action only to the selected item. */
			SINGLE_ITEM
		}

		/** Name of the associated action, i.e. its entry. */
		private String entryName;

		/** {@link AnnotationEntry} elements to apply the associated action to. */
		private ActionScope scope;

		/** {@code instanceKey} with which the action is associated (if any). */
		private String instanceKey;

		/** List of required input parameters to the annotation specific action. */
		private AnnotationActionInputParameters inputParameters;

		/** List of required input parameters to the annotation specific action. */
		private AnnotationActionReqInputParameters inputReqParameters;

		/** Flag if the action performs write (for undo/redo actions). */
		private boolean isWrite;

		/** Constructor defining the name and the scope of the corresponding action. */
		public AnnotationActionEntry(String entryName, ActionScope scope, String instanceKey,
				boolean isWrite) {
			this.entryName = entryName;
			this.scope = scope;
			this.instanceKey = instanceKey;
			this.isWrite = isWrite;
		}

		/**
		 * Constructor defining the name and the scope of the corresponding action. Additionally, a
		 * list of required input parameters types and their display name must be given.
		 */
		public AnnotationActionEntry(String entryName, ActionScope scope, String instanceKey,
				boolean isWrite, AnnotationActionReqInputParameters inputReqParameters) {
			this(entryName, scope, instanceKey, isWrite);
			this.inputReqParameters = inputReqParameters;
		}

		/** Returns the name of the corresponding action. */
		public String getName() {
			return entryName;
		}

		/** Returns the scope of the action, i.e. the set of elements to apply the action to. */
		public ActionScope getScope() {
			return scope;
		}

		/**
		 * Returns the {@code instanceKey} of the corresponding action (multiple actions per
		 * {@link IAnnotatedSpecification} may exist).
		 */
		public String getInstanceKey() {
			return instanceKey;
		}

		/** Returns a list of entered input parameters. */
		public AnnotationActionInputParameters getInputParameters() {
			return inputParameters;
		}

		/** Returns a list of expected input parameters. */
		public AnnotationActionReqInputParameters getRequiredInputParameters() {
			return inputReqParameters;
		}

		/** Returns whether the corresponding action performs a write operation. */
		public boolean isWrite() {
			return isWrite;
		}
	}

	/**
	 * Defines the the parameters required by an annotation action. Further the parameters can be
	 * passed to the action using the same object (this).
	 */
	public class AnnotationActionInputParameters {
		/** References the action-specific parameters and their values. */
		Map<AnnotationActionInputParameter, Object> parameterValueMap = new LinkedHashMap<>();

		/**
		 * Adds the given parameter identifier (Name and Type). Until no associated value is set via
		 * the {@link #put(String, Class, Number)} method, the associated value will be {@code null}
		 * .
		 * 
		 * @param parameterName
		 *            Name of the parameter.
		 * @param parameterType
		 *            Type of the parameter.
		 */
		public void add(String parameterName, Class<?> parameterType) {
			parameterValueMap.put(new AnnotationActionInputParameter(parameterName, parameterType),
					null);
		}

		/**
		 * Adds the value associated with the given parameter description and value type.
		 * 
		 * @param parameterName
		 *            Name of the parameter.
		 * @param parameterType
		 *            Type of the parameter.
		 * @param value
		 *            Value to be added.
		 */
		public <T extends Number> void put(String parameterName, Class<T> parameterType, T value) {
			parameterValueMap.put(new AnnotationActionInputParameter(parameterName, parameterType),
					value);
		}

		/**
		 * Returns the value associated with the given parameter description and value type.
		 * 
		 * @param parameterName
		 *            Name of the parameter.
		 * @param parameterType
		 *            Type of the parameter.
		 * @return the contained value, or {@code null} if none exists.
		 */
		@SuppressWarnings("unchecked")
		public <T> T get(String parameterName, Class<T> parameterType) {
			Object retVal =
					parameterValueMap.get(new AnnotationActionInputParameter(parameterName,
							parameterType));
			if((retVal != null && parameterType.isAssignableFrom(retVal.getClass())) ||
					(retVal instanceof Number && isExpectedType((Number)retVal,
							(Class<? extends Number>)parameterType))) {
				return (T)retVal;
			}
			return null;
		}

		/** Returns the keys of {@code this} AnnotationActionInputParameter. */
		public Set<AnnotationActionInputParameter> keySet() {
			return parameterValueMap.keySet();
		}
	}

	/** Wrapper class to specify a single input parameter of an annotation action. */
	public class AnnotationActionInputParameter extends Pair<String, Class<?>> {
		/** Constructor. */
		public AnnotationActionInputParameter(String first, Class<?> second) {
			super(first, second);
		}
	}

	/**
	 * Defines the required input parameters to execute an action for an
	 * {@link IAnnotatedSpecification} (and its {@code instanceKey}).
	 */
	public class AnnotationActionReqInputParameter extends
			Pair<String, EStructuralFeatureDescriptor> {
		/** Constructor. */
		public AnnotationActionReqInputParameter(String first, EStructuralFeatureDescriptor second) {
			super(first, second);
		}
	}

	/** {@link List} of {@link AnnotationActionReqInputParameter}s. */
	public class AnnotationActionReqInputParameters extends
			ArrayList<AnnotationActionReqInputParameter> {
		// Container.
	}

	/**
	 * Container class to specify the parameters passed to a {@link BiConsumer} executed on
	 * annotations.
	 */
	public class AnnotationActionParameters<T extends IAnnotatedSpecification> {

		/** {@link IAnnotatedSpecification} for which the action was triggered. */
		private T annotation;

		/** List of input parameters for the executing {@link BiConsumer}. */
		private AnnotationActionInputParameters inputParameters;

		/**
		 * Constructor.
		 * 
		 * @param annotation
		 *            {@link IAnnotatedSpecification} for which the action was triggered.
		 * @param inputParameters
		 *            Possibly empty list of input parameters to the executing {@link BiConsumer}.
		 */
		public AnnotationActionParameters(T annotation,
				AnnotationActionInputParameters inputParameters) {
			this.annotation = annotation;
			this.inputParameters = inputParameters;
		}

		/** Returns the {@link IAnnotatedSpecification} for which the action was triggered. */
		public T getAnnotation() {
			return annotation;
		}

		/** Returns the list of input parameters to the executed {@link BiConsumer} */
		public AnnotationActionInputParameters getInputParameters() {
			return inputParameters;
		}
	}

	/** Class to define actions executed on defined annotation. */
	public class AnnotationAction<T extends IAnnotatedSpecification>
			extends
			Pair<AnnotationActionEntry, BiConsumer<AnnotationInstSpec<T>, AnnotationActionParameters<T>>> {
		/** Constructor. */
		public AnnotationAction(AnnotationActionEntry first,
				BiConsumer<AnnotationInstSpec<T>, AnnotationActionParameters<T>> second) {
			super(first, second);
		}
	}

	/**
	 * {@link IAnnotationValueProvider.AnnotationAction}s defined for a specific
	 * {@link IAnnotatedSpecification}.
	 */
	public class AnnotationActions<T extends IAnnotatedSpecification> extends
			ArrayList<AnnotationAction<T>> {
		// Wrapper class.
	}

	/**
	 * Default key used in case this {@link IAnnotationValueProvider} is used to manage a single
	 * instance key only.
	 */
	public static String DEFAULT_KEY = null;

	/** Returns the name of the annotation (e.g., used as labels in annotation views). */
	public String getAnnotationName(T specification);

	/**
	 * Sets a new name for the annotation (e.g., used as labels in annotation views). The return
	 * value indicates if the rename operation was successful.
	 */
	public boolean setAnnotationName(String name, T specification);

	/** Predicate if the the given annotation should be hidden from any view. */
	public boolean isHidden(T specification);

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(T specification);

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(T specification, String instanceKey);

	/** Returns the value of the annotation. */
	public <U> U getAnnotationValue(T specification) throws Exception;

	/** Returns the value of a multi-instance annotation's instance denoted by {@code instanceKey}. */
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception;

	/**
	 * Sets a new value for the annotation.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from the UI internal value representation ({@code U})
	 *             to the respective specialization of {@link IAnnotatedSpecification} has failed,
	 *             or if this operation is not available.
	 */
	public <U> void setAnnotationValue(U value, T specification) throws Exception;

	/**
	 * Sets a new value in the annotation from a {@link String} initializer.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from {@link String} to the respective specialization
	 *             of {@link IAnnotatedSpecification} has failed, or if this operation is not
	 *             available.
	 */
	public void setAnnotationValue(String value, T specification) throws Exception;

	/**
	 * Sets a new value for a multi-instance annotation's instance denoted by {@code instanceKey}.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from the UI internal value representation ({@code U})
	 *             to the respective specialization of {@link IAnnotatedSpecification} has failed,
	 *             or if this operation is not available.
	 */
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception;

	/**
	 * Sets a new value for a multi-instance annotation's instance denoted by {@code instanceKey}
	 * from a {@link String} initializer.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from {@link String} to the respective specialization
	 *             of {@link IAnnotatedSpecification} has failed, or if this operation is not
	 *             available.
	 */
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception;

	/**
	 * Factory method that creates a new {@link IAnnotatedSpecification} specialization of type T to
	 * the given model {@code element} and returns a reference to it.
	 * 
	 * In case the annotation already exists in the model, this method does not create a new object,
	 * but returns the existing one.
	 * 
	 * @throws Exception
	 *             is thrown if no existing element could be retrieved, and the creation of a new
	 *             element has failed.
	 */
	public T getAnnotatedSpecificationForModelElement(IModelElement element) throws Exception;

	/**
	 * Returns the concrete specialization {@code T} of {@link IAnnotatedSpecification} supported by
	 * this specialization of {@link IAnnotationValueProvider}.
	 */
	public Class<T> getAnnotationClazz();

	/**
	 * Returns a {@link List} of model element types which will not have the
	 * {@link IAnnotatedSpecification} managed by this {@link IAnnotationValueProvider}, refining
	 * the registration of the respective {@link IAnnotatedSpecification} in the
	 * {@code org.fortiss.tooling.base.ui.annotation} extension point.
	 */
	public List<Class<? extends EObject>> excludeModelElementTypeFromAnnotatedSpecification();

	/**
	 * Returns the instance keys supported by this {@link IAnnotationValueProvider}.
	 * {@link IAnnotationValueProvider}s can either define a static {@link List}, or enable the
	 * dynamic creation of instance keys using {@link #allowsDynamicAnnotationInstances()}. In the
	 * latter case, this method returns current list of dynamic annotation instances.
	 */
	public Collection<String> getInstanceKeys(T specification);

	/** Returns the {@link EStructuralFeatureDescriptor} for a given {@code instanceKey}. */
	public EStructuralFeatureDescriptor getEStructuralFeatureDescriptor(String instanceKey);

	/**
	 * <p>
	 * Predicate whether the {@link IAnnotatedSpecification} managed by this
	 * {@link IAnnotationValueProvider} contains a value type allowing the storage of dynamically
	 * created instances of the annotation (like a list of strings).
	 * </p>
	 * <p>
	 * Static instances can be implemented by overriding
	 * {@link #getInstanceKeys(IAnnotatedSpecification)} and returning the {@link List} of
	 * admissible instance keys.
	 * </p>
	 */
	public boolean allowsDynamicAnnotationInstances();

	/**
	 * <p>
	 * Returns a list of {@link AnnotationActionEntry}s <--> {@link BiConsumer} {@link Pair}s that
	 * define the context menu entries that are available for specific entries of an
	 * {@link IAnnotatedSpecification}.
	 * </p>
	 * <p>
	 * This list is used in the view part of the annotations to provide actions for concrete
	 * {@link AnnotationEntry}s that can be triggered via the context menu in the annotation view.
	 * The action is applied to the subset of {@link AnnotationEntry}s provided by the underlying
	 * annotation view (typically, the set of visible annotations that match the current filters).
	 * </p>
	 * 
	 * @return List of AnnotationActionEntry <--> {@link BiConsumer} {@link Pair}s that define the
	 *         context menu entries that are available for specific entries of an
	 *         {@link IAnnotatedSpecification}.
	 */
	AnnotationActions<T> getContextMenuEntries(T specification);
}
