/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.compose;

import static org.eclipse.emf.ecore.util.EcoreUtil.delete;

import java.util.ArrayList;
import java.util.List;
import java.util.Observable;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementReference;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.kernel.extension.IElementCompositor;
import org.fortiss.tooling.kernel.extension.data.IElementCompositionContext;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Base implementation of a compositor for {@link EObject}s which is responsible for the standard
 * composition of the {#link contained} {@link EObject}.
 * 
 * @author diewald
 */
public abstract class ModelElementCompositorBase<C extends EObject> extends Observable implements
		IElementCompositor<C> {

	/**
	 * {@inheritDoc}
	 * <p>
	 * <b>Note: Derived compositors must call this method.</b>
	 * </p>
	 */
	@Override
	public boolean compose(C container, EObject contained, IElementCompositionContext context) {
		if(contained instanceof IModelElement) {
			// Instantiate annotations for new model element. Derived compositors must call this
			// method
			// (see #2228 and #2208)
			IAnnotationValueService.getInstance().instantiateAnnotations((IModelElement)contained);

			// Assign IDs for newly created annotations. Notes:
			// - Do this for each specification instead of the "contained" model element. Since
			// "contained" might already have an any, this would trigger an expensive algorithm to
			// check if there are duplicated IDs.
			// - The assignment of IDs cannot be performed in instantiateAnnotations() above since
			// "contained" is not hooked to the model yet. Instead, the "container" is used as a
			// reference to assign the annotations' IDs.
			final ITopLevelElement topLevelElement =
					IPersistencyService.getInstance().getTopLevelElementFor(container);
			for(IModelElementSpecification specification : ((IModelElement)contained)
					.getSpecifications()) {
				topLevelElement.prepareIDs(specification);
			}
		}
		makeNameDistinct(container, contained);
		return true;
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Decomposition that takes care of deleting {@link IModelElement} specific references etc..
	 * </p>
	 * <p>
	 * <b>Note: Derived compositors MAY NOT CALL THIS METHOD via {@code super.decompose(...)}!.</b>
	 * </p>
	 */
	@Override
	public boolean decompose(EObject contained) {
		// Removes first the constraints
		if(contained instanceof IConstrained) {
			for(ConstraintInstance ci : ((IConstrained)contained).getConstraintInstances()) {
				for(IConstrained c : ci.getConstraineds()) {
					c.getConstraintInstances().remove(ci);
				}
				EcoreUtil.delete(ci, true);
			}
		}
		return iterateDecompose(contained) && decomposeSpecific(contained);
	}

	/**
	 * Base implementation for element removal using EcoreUtil.delete
	 * (non-recursively). Subclasses may override.
	 */
	protected boolean decomposeSpecific(EObject contained) {
		EcoreUtil.delete(contained, true);
		return true;
	}

	/** Recursive traversal of all model elements in subtree. */
	protected boolean iterateDecompose(EObject contained) {
		if(contained instanceof IModelElement) {
			if(!iterateDecomposeSpecifications((IModelElement)contained)) {
				return false;
			}
			if(!iterateDecomposeReference((IModelElement)contained)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over references of given {@link IModelElement} and performs
	 * decomposition.
	 */
	private boolean iterateDecomposeReference(IModelElement me) {
		ArrayList<IModelElementReference> list =
				new ArrayList<IModelElementReference>(me.getReferencedBy());
		for(IModelElementReference ref : list) {
			if(!decomposeReference(ref)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over specifications of given {@link IModelElement} and performs
	 * decomposition.
	 */
	private boolean iterateDecomposeSpecifications(IModelElement me) {
		ArrayList<IModelElementSpecification> list =
				new ArrayList<IModelElementSpecification>(me.getSpecifications());
		for(IModelElementSpecification spec : list) {
			if(!decomposeSpecification(spec)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Base implementation uses element compositor service's decomposition by
	 * default and returns <code>true</code>. Subclasses may override.
	 */
	protected boolean decomposeSpecification(IModelElementSpecification element) {
		if(IElementCompositorService.getInstance().canDecompose(element)) {
			return IElementCompositorService.getInstance().decompose(element);
		}
		// no delete here, since normally specs are destroyed with the element
		return true;
	}

	/**
	 * Base implementation uses element compositor service's decomposition by
	 * default and returns <code>true</code>. Subclasses may override.
	 */
	protected boolean decomposeReference(IModelElementReference element) {
		element.setReference(null);
		delete(element, true);
		return true;
	}

	/**
	 * This function checks if the name of any element of same type and name as
	 * <code>contained</code> exists in <code>container</code>, and modifies the name in case if
	 * required.
	 */
	public void makeNameDistinct(EObject container, EObject contained) {
		if(!(contained instanceof INamedElement) || ((INamedElement)contained).getName() == null) {
			return;
		}
		INamedElement neNew = (INamedElement)contained;
		String newName = ((INamedElement)contained).getName();
		List<String> listNames = new ArrayList<String>();

		for(EObject o : container.eContents()) {
			if(o.getClass().isInstance(contained) && o instanceof INamedElement && o != contained) {
				listNames.add(((INamedElement)o).getName());
			}
		}

		if(!listNames.contains(newName)) {
			return;
		}

		int index = 1;
		while(listNames.contains(newName + index)) {
			index++;
		}

		neNew.setName(neNew.getName() + index);
	}
}
