/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.ParsePosition;
import java.util.Locale;

import org.apache.commons.lang.math.NumberUtils;

/**
 * Utility class to support mathematical operations.
 * 
 * @author diewald
 */
public class BaseMathUtils {

	/**
	 * Converts the given {@code number} to the given {@code targetType}, possibly throwing an
	 * {@link ArithmeticException} if an over- or underflow occurs during conversion.
	 * If a floating point number is given and shall be transformed to a integer-kind number, a
	 * rounding operation is performed.
	 * 
	 * NOTE: If the conversion fails, an {@link ArithmeticException} will be thrown.
	 * 
	 * @param number
	 *            {@link Number} to be converted.
	 * @param targetType
	 *            Type of the converted {@link Number}
	 * @return converted {@link Number}.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Number> T convertNumber(Number number, Class<T> targetType) {
		if(targetType.isInstance(number)) {
			return (T)number;
		}

		// Integer-kind numbers.
		// - Rounding to long in case of a floating point number input.
		// - Bound checking if needed.
		// - Conversion.
		else if(targetType.equals(Byte.class) || targetType.equals(Byte.TYPE)) {
			long value = Long.MIN_VALUE;
			if(number instanceof Float || number instanceof Double) {
				value = Math.round(number.doubleValue());
			} else {
				value = number.longValue();
			}
			if(value < Byte.MIN_VALUE || value > Byte.MAX_VALUE) {
				throw createOverUnderFlowException(number, targetType);
			}
			return (T)Byte.valueOf(number.byteValue());
		} else if(targetType.equals(Short.class) || targetType.equals(Short.TYPE)) {
			long value = Long.MIN_VALUE;
			if(number instanceof Float || number instanceof Double) {
				value = Math.round(number.doubleValue());
			} else {
				value = number.longValue();
			}
			if(value < Short.MIN_VALUE || value > Short.MAX_VALUE) {
				throw createOverUnderFlowException(number, targetType);
			}
			return (T)Short.valueOf(number.shortValue());
		} else if(targetType.equals(Integer.class) || targetType.equals(Integer.TYPE)) {
			long value = Long.MIN_VALUE;
			if(number instanceof Float || number instanceof Double) {
				value = Math.round(number.doubleValue());
			} else {
				value = number.longValue();
			}
			if(value < Integer.MIN_VALUE || value > Integer.MAX_VALUE) {
				createOverUnderFlowException(number, targetType);
			}
			return (T)Integer.valueOf(number.intValue());
		} else if(targetType.equals(Long.class) || targetType.equals(Long.TYPE)) {
			long value = Long.MIN_VALUE;
			if(number instanceof Float || number instanceof Double) {
				value = Math.round(number.doubleValue());
			} else {
				value = number.longValue();
			}
			return (T)Long.valueOf(value);
		} else if(targetType.equals(BigInteger.class)) {
			return (T)BigInteger.valueOf(number.longValue());
		}

		// Floating point numbers.
		// - Bound checking if needed.
		// - Conversion.
		else if(targetType.equals(Float.class) || targetType.equals(Float.TYPE)) {
			double value = number.doubleValue();
			if(value < Float.MIN_VALUE || value > Float.MAX_VALUE) {
				throw createOverUnderFlowException(number, targetType);
			}
			return (T)Float.valueOf(number.floatValue());
		} else if(targetType.equals(Double.class) || targetType.equals(Double.TYPE)) {
			return (T)Double.valueOf(number.doubleValue());
		} else if(targetType.equals(BigDecimal.class)) {
			return (T)new BigDecimal(number.doubleValue());
		}

		throw new ArithmeticException("Unable to convert the number " + number + " to the type " +
				targetType.getSimpleName() + ".");
	}

	/**
	 * Returns the {@link Number} of the given {@code numString} if the string is a) a number and b)
	 * has the given {@code expectedType}. If either condition is violated, {@code null} is
	 * returned.
	 * 
	 * @param numString
	 *            String of a number.
	 * @param expectedType
	 *            Expected number type of the given {@code numString}.
	 * @return {@link Number} derived from the given {@code numString}.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Number> T getNumberFromString(String numString, Class<T> expectedType) {
		Number value = null;
		if(NumberUtils.isNumber(numString)) {
			if(expectedType.isAssignableFrom(Byte.class) || expectedType.equals(Byte.TYPE)) {
				value = Byte.valueOf(numString);
			} else if(expectedType.isAssignableFrom(Double.class) ||
					expectedType.equals(Double.TYPE)) {
				value = Double.valueOf(numString);
			} else if(expectedType.isAssignableFrom(Float.class) || expectedType.equals(Float.TYPE)) {
				value = Float.valueOf(numString);
			} else if(expectedType.isAssignableFrom(Integer.class) ||
					expectedType.equals(Integer.TYPE)) {
				value = Integer.valueOf(numString);
			} else if(expectedType.isAssignableFrom(Long.class) || expectedType.equals(Long.TYPE)) {
				value = Long.valueOf(numString);
			} else if(expectedType.isAssignableFrom(Short.class) || expectedType.equals(Short.TYPE)) {
				value = Short.valueOf(numString);
			} else if(expectedType.isAssignableFrom(BigDecimal.class)) {
				// We always have US-style numbers.
				DecimalFormat df = (DecimalFormat)NumberFormat.getInstance(new Locale("en", "US"));
				df.setParseBigDecimal(true);
				value = df.parse(numString, new ParsePosition(0));
			} else if(expectedType.isAssignableFrom(BigInteger.class)) {
				value = new BigInteger(numString);
			}
		}

		return (T)value;
	}

	/**
	 * Returns whether the given {@code value} has the given type. This method respects primitive
	 * types as the {@code expectedType}s.
	 * 
	 * @param value
	 *            to be checked.
	 * @param expectedType
	 *            target type (including primitive types).
	 * @return see above.
	 */
	public static boolean isExpectedType(Number value, Class<? extends Number> expectedType) {
		if(expectedType.isAssignableFrom(value.getClass())) {
			return true;
		} else if(value instanceof Byte && expectedType.equals(Byte.TYPE)) {
			return true;
		} else if(value instanceof Double && expectedType.equals(Double.TYPE)) {
			return true;
		} else if(value instanceof Float && expectedType.equals(Float.TYPE)) {
			return true;
		} else if(value instanceof Integer && expectedType.equals(Integer.TYPE)) {
			return true;
		} else if(value instanceof Long && expectedType.equals(Long.TYPE)) {
			return true;
		} else if(value instanceof Short && expectedType.equals(Short.TYPE)) {
			return true;
		}
		return false;
	}

	/**
	 * Predicate whether the given Number is an integer of the types Byte, Short, Integer, Long,
	 * BigInteger.
	 * 
	 * @param value
	 *            Number to test.
	 * @return boolean if the given Number is an integer number.
	 */
	public static boolean isIntegerNumber(Number value) {
		return(value instanceof Byte || value instanceof Short || value instanceof Integer ||
				value instanceof Long || value instanceof BigInteger);
	}

	/**
	 * Creates an {@link ArithmeticException} if an over-/underflow occurs when converting the given
	 * {@code number} to the given {@code targetType}.
	 */
	public static ArithmeticException createOverUnderFlowException(Number number,
			Class<?> targetType) {
		return new ArithmeticException("Number Over-/Underflow deetected: The number " + number +
				" is too large to be casted to the type " + targetType.getSimpleName() + ".");
	}
}
