/*-------------------------------------------------------------------------+
| Copyright 2012 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.graphicsGL.ui.objects;

import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.createBillboardMatrix;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.drawLine;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.drawMaterialGroups;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.flushName;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.moveTo;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.multiplyCurrentMatrixWith;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.normalize;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.prepareName;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.restoreMatrix;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.rotate;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.saveMatrix;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.scale;
import static org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives.setColor;
import static org.fortiss.tooling.graphicsGL.ui.util.Graph3DUtil.convert;
import static org.fortiss.tooling.graphicsGL.ui.util.Graph3DUtil.importCharacters;
import static org.fortiss.tooling.graphicsGL.ui.util.Graph3DUtil.importLetters;
import static org.fortiss.tooling.graphicsGL.ui.util.Graph3DUtil.importNumbers;
import static org.fortiss.tooling.graphicsGL.util.Graphics3DModelElementFactory.BLACK;
import static org.fortiss.tooling.graphicsGL.util.Graphics3DModelElementFactory.LIGHT_BLUE;
import static org.fortiss.tooling.graphicsGL.util.Graphics3DModelElementFactory.WHITE;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.fortiss.tooling.graphicsGL.model.Color3D;
import org.fortiss.tooling.graphicsGL.model.graph.Axis3D;
import org.fortiss.tooling.graphicsGL.model.graph.CoordinateSystem3D;
import org.fortiss.tooling.graphicsGL.model.graph.Graph3D;
import org.fortiss.tooling.graphicsGL.model.graph.GraphLine3D;
import org.fortiss.tooling.graphicsGL.model.graph.GraphPoint3D;
import org.fortiss.tooling.graphicsGL.model.graph.GraphText3D;
import org.fortiss.tooling.graphicsGL.model.scene.Object3D;
import org.fortiss.tooling.graphicsGL.ui.camera.Camera;
import org.fortiss.tooling.graphicsGL.ui.picker.ModelObjectPicker;
import org.fortiss.tooling.graphicsGL.ui.util.GLPrimitives;
import org.lwjglx.util.glu.GLU;
import org.lwjglx.util.glu.Sphere;
import org.lwjglx.util.vector.Vector3f;
import org.lwjglx.util.vector.Vector4f;

/**
 * View object for {@link Graph3D}.
 * 
 * @author hoelzl
 */
public class Graph3DViewObject extends ViewObjectBase {
	/** The graph model. */
	private final Graph3D graph;

	/** The highlighted model object. */
	private GraphPoint3D highlighted;

	/** the distance between letters in a text */
	static float DISTANCE_BEWTWEEN_LETTERS = 0.7f;

	/** Hash map of all letters which can be rendered */
	HashMap<String, Object3D> letters;
	/** Hash map of all characters which can be rendered */
	HashMap<String, Object3D> characters;
	/** Hash map of all numbers which can be rendered */
	HashMap<Integer, Object3D> numbers;

	/** Constructor. */
	public Graph3DViewObject(Graph3D graph) {
		this.graph = graph;
		letters = importLetters();
		characters = importCharacters();
		numbers = importNumbers();
	}

	/** Sets the currently highlighted point. */
	public void setHighlightedPoint(GraphPoint3D point) {
		highlighted = point;
	}

	/** {@inheritDoc} */
	@Override
	public void renderObject(Camera c) {
		// GLPrimitives.scale(1.5f, 1.5f, 1.5f);
		renderCoordSys(graph.getCoordinateSystem());
		setColor(WHITE);
		for(GraphPoint3D p : graph.getPoints()) {
			if(p == highlighted) {
				setColor(LIGHT_BLUE);
				renderPoint(p);
			} else {
				setColor(WHITE);
				Color3D color = p.getColor();
				if(color != null) {
					setColor(color);
				}
				renderPoint(p);
				setColor(WHITE);
			}
		}

		setColor(BLACK);
		for(GraphLine3D l : graph.getLines()) {
			Color3D co = l.getColor();
			if(co != null)
				setColor(co);
			drawLine(l.getStart().getPosition(), l.getEnd().getPosition());
			setColor(BLACK);
		}

		setColor(BLACK);
		for(GraphText3D t : graph.getTexts()) {
			renderText(c, t.getText(), convert(t.getPosition()));
		}

		setColor(WHITE);

	}

	/** Renders objects as billboards at the given position */
	private void renderText(Camera camera, String string, Vector3f pos) {
		List<Object3D> toRender = new ArrayList<Object3D>();
		for(int i = 0; i < string.length(); ++i) {
			String st = string.charAt(i) + "";

			Object3D renderObject = this.letters.get(st.toUpperCase());
			// check if it is a number
			if(renderObject == null) {
				try {
					renderObject = this.numbers.get(Integer.decode(st));
				} catch(NumberFormatException e) {
					renderObject = null;
				}
			}
			// if it is neither a letter, number, white space or "_" print a "?"
			if(renderObject == null) {
				if(!st.equals(" ") && !st.equals("_"))
					renderObject = this.characters.get("?");
			}
			toRender.add(renderObject);
		}

		renderText(camera, toRender, pos);

	}

	/** Renders objects as billboards at the given position */
	private void renderText(Camera c, List<Object3D> objects, Vector3f pos) {

		saveMatrix();

		Vector3f look = new Vector3f();
		// Vector3f pos = new Vector3f(0, 10, 0);
		Vector3f right = new Vector3f();
		Vector3f up;

		// look is now the vector from the camera position to the text position
		// billboarding goes along z axis
		Vector3f.sub(c.getPosition(), pos, look);
		look.z = 0;
		look = normalize(look);

		up = new Vector3f(0, 0, 1);
		Vector3f.cross(up, look, right);

		// set up transformation matrix for text / billboard
		float[] bbmat = new float[16];
		createBillboardMatrix(bbmat, right, up, look, pos);

		// apply the billboard
		multiplyCurrentMatrixWith(bbmat);

		// rotate text so it is readable from left to right
		rotate(new Vector4f(0, 0, 1, -90));

		scale(1.75f, 1.75f, 1.75f);

		// draw the Text
		for(Object3D o : objects) {
			if(o != null)
				drawMaterialGroups(o.getMesh().getMaterialGroups());
			GLPrimitives.translate(0, DISTANCE_BEWTWEEN_LETTERS, 0);
		}

		restoreMatrix();

	}

	/** {@inheritDoc} */
	@Override
	public void renderObjectForPicking(ModelObjectPicker picker, Camera c) {
		// only points are pickable
		for(GraphPoint3D p : graph.getPoints()) {
			prepareName(p, picker);
			renderPoint(p);
			flushName();
		}
	}

	/** The GLU sphere used for {@link GraphPoint3D} rendering. */
	private static Sphere SPHERE = new Sphere();

	static {
		SPHERE.setDrawStyle(GLU.GLU_FILL);
	}

	/** Renders the given point. */
	private void renderPoint(GraphPoint3D p) {
		saveMatrix();
		moveTo(p.getPosition());
		scale(1.5f, 1.5f, 1.5f);
		SPHERE.draw(0.2f, 9, 4);
		restoreMatrix();
	}

	/** Renders the coordinate system if present. */
	private void renderCoordSys(CoordinateSystem3D coordinateSystem) {
		if(coordinateSystem != null) {
			if(coordinateSystem.getXAxis() != null) {
				Axis3D a = coordinateSystem.getXAxis();
				renderAxis(a.getColor(), a.getLength(), 0, 0);
			}
			if(coordinateSystem.getYAxis() != null) {
				Axis3D a = coordinateSystem.getYAxis();
				renderAxis(a.getColor(), 0, a.getLength(), 0);
			}
			if(coordinateSystem.getZAxis() != null) {
				Axis3D a = coordinateSystem.getZAxis();
				renderAxis(a.getColor(), 0, 0, a.getLength());
			}
		}
	}

	/** Renders the axis. */
	private void renderAxis(Color3D color, float x, float y, float z) {
		setColor(color);
		drawLine(0, 0, 0, x, y, z);
	}
}
