/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal;

import static java.util.Collections.emptyList;

import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.EditPart;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.fortiss.tooling.kernel.ui.ToolingKernelUIActivator;
import org.fortiss.tooling.kernel.ui.extension.IEditPartFactory;
import org.fortiss.tooling.kernel.ui.internal.introspection.items.EditPartFactoryServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.ui.service.IEditPartFactoryService;
import org.fortiss.tooling.kernel.utils.LoggingUtils;

/**
 * This class implements the {@link IEditPartFactoryService} interface.
 * 
 * @author hoelzl
 */
public class EditPartFactoryService extends EObjectAwareServiceBase<IEditPartFactory> implements
		IEditPartFactoryService, IIntrospectiveKernelService {
	/** The singleton service instance. */
	private static EditPartFactoryService INSTANCE = new EditPartFactoryService();

	/** Returns the service instance. */
	public static EditPartFactoryService getInstance() {
		return INSTANCE;
	}

	/** The prototype provider extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.ui.editPartFactory";

	/** The prototype provider configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "editPartFactory";

	/** The model element handler class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "factory";

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** Registers the given editpart factory with the service. */
	@Override
	public void registerEditpartFactory(IEditPartFactory factory, Class<EObject> modelElementClass) {
		addHandler(modelElementClass, factory);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service is used to register edit part factories for graphical editors." +
				"\nEach factory transforms model elements into graphical edit parts as defined by the GEF framework." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public EditPart createEditPart(EditPart context, Object modelElement) {

		if(!(modelElement instanceof EObject)) {
			LoggingUtils.error(ToolingKernelUIActivator.getDefault(),
					EditPartFactoryService.class.getSimpleName() +
							" may only be used for EObjects!");
			return null;
		}
		Class<? extends EObject> clazz = ((EObject)modelElement).getClass();

		List<IEditPartFactory> factories = getRegisteredHandlers(clazz);
		if(factories == null || factories.isEmpty()) {
			LoggingUtils.error(ToolingKernelUIActivator.getDefault(),
					"No editpart factory registered for " + clazz.getName());
			return null;
		}

		for(IEditPartFactory factory : factories) {
			EditPart part = factory.createEditPart(context, modelElement);
			if(part != null) {
				return part;
			}
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Editpart Factory Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new EditPartFactoryServiceIntrospectionDetailsItem(handlerMap);
	}
}
