/*-----------------------------------------------------------------------+
 | org.fortiss.tooling.kernel.ui.internal.views
 |                                                                       |
   $Id$            
 |                                                                       |
 | Copyright (c)  2016-today fortiss GmbH                                |
 |                                                                       |
 +-----------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal.views;

import static java.util.Collections.emptyList;
import static org.eclipse.jface.resource.ImageDescriptor.createFromImage;

import java.util.List;
import java.util.stream.Collectors;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.DecorationOverlayIcon;
import org.eclipse.jface.viewers.IDecoration;
import org.eclipse.swt.graphics.Image;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.ErrorConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.FailedConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.SuccessConstraintInstanceStatus;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.ui.ESharedImages;
import org.fortiss.tooling.kernel.ui.ToolingKernelUIActivator;
import org.fortiss.tooling.kernel.ui.extension.IContextMenuContributor;
import org.fortiss.tooling.kernel.ui.extension.IModelElementHandler;
import org.fortiss.tooling.kernel.ui.extension.data.ContextMenuContextProvider;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService;
import org.fortiss.tooling.kernel.ui.service.IConstraintUIService.IFix;
import org.fortiss.tooling.kernel.ui.service.IContextMenuService;
import org.fortiss.tooling.kernel.ui.service.IModelEditorBindingService;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

/**
 * Context menu for constraints.
 * 
 * @author hoelzl
 */
public class ConstraintMenu implements IContextMenuContributor {

	/** {@inheritDoc} */
	@Override
	public List<IContributionItem> getContributedItems(EObject selectedObject,
			ContextMenuContextProvider contextProvider) {
		if(!(selectedObject instanceof IConstrained)) {
			return emptyList();
		}
		return ((IConstrained)selectedObject).getConstraintInstances().stream()
				.filter(ci -> shouldGetMenuEntry(ci))
				.map(ci -> constraintToAction(ci, (IConstrained)selectedObject))
				.collect(Collectors.toList());
	}

	/**
	 * <code>true</code> iff <code>ci</code> should get a menu entry, i.e., if not successful and
	 * activated.
	 */
	private boolean shouldGetMenuEntry(ConstraintInstance ci) {
		IConstraintInstanceStatus status = IConstraintUIService.getInstance().getStatus(ci);
		return !(status == null || status instanceof SuccessConstraintInstanceStatus);
	}

	/**
	 * Returns the submenu or menu entry corresponding to the status of <code>ci</code>.
	 * <code>elt</code> is needed to prevent displaying an entry to go to <code>elt</code> (not
	 * hurtful, but useless, since the user is already seeing it).
	 */
	private IContributionItem constraintToAction(ConstraintInstance ci, IConstrained elt) {
		IConstraintInstanceStatus status = IConstraintUIService.getInstance().getStatus(ci);
		ConstraintSubMenuBase m;
		if(status instanceof FailedConstraintInstanceStatus) {
			m = new KnowMoreAboutFailingConstraintAction(ci, elt);
		} else if(status instanceof ErrorConstraintInstanceStatus) {
			m = new CheckErrorConstraintAction(ci, elt);
		} else if(status instanceof OutdatedConstraintInstanceStatus) {
			m = new CheckOutdatedConstraintAction(ci, elt);
		} else {
			// should not happen
			return null;
		}
		if(m.getItems().length == 0) {
			Action a = new Action(m.getMenuText(), m.getImageDescriptor()) {
				// nothing to do
			};
			return new ActionContributionItem(a);
		}
		if(m.getItems().length == 1 && m.getTopActionContribution() != null) {
			IAction uniqueAction = m.getTopActionContribution().getAction();
			String txt = m.getMenuText();
			String newText = txt + " -> " + uniqueAction.getText();
			uniqueAction.setText(newText);
			uniqueAction.setImageDescriptor(m.getImageDescriptor());
			return m.getTopActionContribution();
		}
		return m;
	}

	/** {@inheritDoc} */
	@Override
	public String getMenuSectionID() {
		return IContextMenuService.BEFORE_GLOBAL_MENU_SECTION_ID;
	}

	/** Action to "go to" a given constrained element. */
	private static class GoToConstrained extends Action {

		/** The constrained element to open. */
		private IConstrained cstrd;

		/** Constructor. */
		public GoToConstrained(IConstrained cstrd) {
			super("Go to " + IModelElementHandlerService.getInstance().getName(cstrd),
					createFromImage(IModelElementHandlerService.getInstance().getIcon(cstrd)));
			this.cstrd = cstrd;
		}

		/** {@inheritDoc} */
		@Override
		public void run() {
			IModelElementHandler<EObject> handler =
					IModelElementHandlerService.getInstance().getModelElementHandler(cstrd);
			EObject eltToOpen =
					handler == null ? cstrd : handler.handleOpenModelElementRequest(cstrd);
			IModelEditorBindingService.getInstance().openInEditor(eltToOpen);
		}
	}

	/** Get the icon of the prototype. */
	public static String getText(ConstraintInstance ci, String suffix) {
		return "\"" + IConstraintUIService.getInstance().getDescription(ci) + "\": " + suffix;
	}

	/** Get the icon of the prototype. */
	public static ImageDescriptor getIcon(ConstraintInstance ci, ImageDescriptor overlay) {
		ImageDescriptor imgd = IConstraintUIService.getInstance().getIconImageDescriptor(ci);
		Image img = imgd == null ? null : imgd.createImage();
		if(img == null) {
			if(IConstraintUIService.getInstance().shallDisplayAsWarning(ci)) {
				return ESharedImages.WARNING.getImageDescriptor();
			}
			return ESharedImages.ERROR.getImageDescriptor();
		}
		ImageDescriptor[] descriptors = new ImageDescriptor[5];
		descriptors[IDecoration.BOTTOM_LEFT] = overlay;
		return new DecorationOverlayIcon(img, descriptors);
	}

	/** Get the correct overlay for the given constraint. */
	private static ImageDescriptor getOverlay(ConstraintInstance ci) {
		if(IConstraintUIService.getInstance().shallDisplayAsWarning(ci)) {
			return ESharedImages.WARNING_OVERLAY.getImageDescriptor();
		}
		return ESharedImages.ERROR_OVERLAY.getImageDescriptor();
	}

	/** Base for action involved in a constraint submenu. */
	private static class ConstraintSubMenuBase extends MenuManager {
		/** The constraint. */
		protected ConstraintInstance ci;

		/**
		 * The action to get more information about the constraint status.
		 * The same action allows to update an outdated constraint status (there is no
		 * "more information" to get out of an outdated constraint); it is stored in a field to
		 * allow the "update" action to change the icon and the text going with it.
		 */
		protected ActionContributionItem moreInfoAction;

		/** Constructor. */
		public ConstraintSubMenuBase(ConstraintInstance ci, IConstrained selectedElt,
				String suffix, ImageDescriptor overlay) {
			super(getText(ci, suffix), getIcon(ci, overlay), null);
			this.ci = ci;
			if(IConstraintUIService.getInstance().canOpen(ci)) {
				moreInfoAction = new ActionContributionItem(new OpenStatusAction());
				this.add(moreInfoAction);
			}
			for(IConstrained cstrd : ci.getConstraineds()) {
				if(!selectedElt.equals(cstrd)) {
					this.add(new ActionContributionItem(new GoToConstrained(cstrd)));
				}
			}
		}

		/**
		 * @return The top action contribution. Useful in case we get aware that there is only one
		 *         contribution and we therefore want to present it without going through a submenu.
		 */
		public ActionContributionItem getTopActionContribution() {
			if(this.getItems()[0] instanceof ActionContributionItem) {
				return (ActionContributionItem)this.getItems()[0];
			}
			return null;
		}

		/** Action to update a constraint. Assumes the status *can be open*. */
		protected class OpenStatusAction extends Action {

			/** Constructor. */
			public OpenStatusAction() {
				super("More information...", ToolingKernelUIActivator
						.getImageDescriptor("icons/info.gif"));
			}

			/** {@inheritDoc} */
			@Override
			public void run() {
				IConstraintInstanceStatus status = IConstraintUIService.getInstance().getStatus(ci);
				if(status instanceof OutdatedConstraintInstanceStatus) {
					ICommandStackService.getInstance().runAsCommand(ci, new Runnable() {
						@Override
						public void run() {
							IConstraintUIService.getInstance().openStatus(ci);
						}
					});
				} else {
					IConstraintUIService.getInstance().openStatus(ci);
				}
			}
		}
	}

	/** Action for creating a new prototype. */
	private static class CheckOutdatedConstraintAction extends ConstraintSubMenuBase {

		/** Constructor. */
		public CheckOutdatedConstraintAction(ConstraintInstance ci, IConstrained selectedElt) {
			super(ci, selectedElt, "Outdated", ESharedImages.WARNING_OVERLAY.getImageDescriptor());
			IAction action = moreInfoAction.getAction();
			action.setImageDescriptor(ToolingKernelUIActivator.getImageDescriptor("icons/ok.png"));
			action.setText("Check");
		}
	}

	/** Action for knowing more about some unsuccessful constraint. */
	private static class KnowMoreAboutUnsuccessfulConstraintAction extends ConstraintSubMenuBase {

		/** Constructor. */
		public KnowMoreAboutUnsuccessfulConstraintAction(ConstraintInstance ci,
				IConstrained selectedElt, String suffix) {
			super(ci, selectedElt, suffix, getOverlay(ci));
			List<IFix> fixes = IConstraintUIService.getInstance().fixes(ci);
			if(fixes != null) {
				for(IFix fix : fixes) {
					this.add(new ActionContributionItem(new FixAction(fix)));
				}
			}
		}

		/** Action to update a constraint. */
		private class FixAction extends Action {

			/** The fix to achieve. */
			IFix fix;

			/** Constructor. */
			public FixAction(IFix fix) {
				super(fix.getDescription(), ToolingKernelUIActivator
						.getImageDescriptor("icons/fix.png"));
				this.fix = fix;
			}

			/** {@inheritDoc} */
			@Override
			public void run() {
				ICommandStackService.getInstance().runAsCommand(ci, new Runnable() {
					@Override
					public void run() {
						fix.runFix(IConstraintUIService.getInstance().getStatus(ci));
					}
				});
			}
		}
	}

	/** Action for knowing more about a failing constraint. */
	private static class KnowMoreAboutFailingConstraintAction extends
			KnowMoreAboutUnsuccessfulConstraintAction {

		/** Constructor. */
		public KnowMoreAboutFailingConstraintAction(ConstraintInstance ci, IConstrained selectedElt) {
			super(ci, selectedElt, "Unsatisfied");
		}
	}

	/** Action for knowing more about a constraint yielding an error. */
	private static class CheckErrorConstraintAction extends
			KnowMoreAboutUnsuccessfulConstraintAction {

		/** Constructor. */
		public CheckErrorConstraintAction(ConstraintInstance ci, IConstrained selectedElt) {
			super(ci, selectedElt, "Error encountered");
		}
	}
}
