/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.extension;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.data.IConnectionCompositionContext;
import org.fortiss.tooling.kernel.service.IConnectionCompositorService;
import org.fortiss.tooling.kernel.service.base.IEObjectAware2;

/**
 * Interface for connection compositor extensions, i.e. classes which connect
 * and disconnect model elements. This may either be a direct connection or by
 * using an additional connection element (which is then composed with a parent
 * element).
 * <P>
 * Connection compositor extensions are handled by {@link IConnectionCompositorService}.
 * <P>
 * A typical example is the creation and removal of an edge in a graph editor. The connection
 * compositor links two nodes of the graph and aggregates the edge with the graph as its parent
 * element.
 * 
 * @param <P>
 *            the meta-model class of the parent elements handled by this
 *            compositor.
 * @param <S>
 *            the meta-model class of source model elements handled by this
 *            compositor.
 * @param <T>
 *            the meta-model class of target model elements handled by this
 *            compositor.
 * 
 * @author hoelzl
 */
public interface IConnectionCompositor<P extends EObject, S extends EObject, T extends EObject>
		extends IEObjectAware2<S, T> {
	/**
	 * Returns whether a connection between the source and the target is
	 * possible for the given parent element. Additional information can be
	 * included in the context, which also may be null.
	 */
	boolean canConnect(S source, T target, P parent, IConnectionCompositionContext context);

	/**
	 * Actually connects the given source and target. This will only be called
	 * by the service if
	 * {@link #canConnect(EObject, EObject, EObject, IConnectionCompositionContext)} returned
	 * <code>true</code>. This method returns a boolean, since
	 * user-interaction during the connect might cancel the connect.
	 * 
	 * @param source
	 *            the source of the connection.
	 * @param target
	 *            the target of the connection.
	 * @param parent
	 *            the parent of the connection.
	 * @param context
	 *            context information for the connection. May be null.
	 * 
	 * @return {@code true} if the connect operation has been performed, {@code false} otherwise
	 *         (e.g., the user has cancelled).
	 */
	boolean connect(S source, T target, P parent, IConnectionCompositionContext context);

	/**
	 * Returns whether the given connection can be altered to connect the given
	 * source and target. Additional information can be included in the context,
	 * which also may be null.
	 */
	boolean canReconnect(EObject connection, S source, T target,
			IConnectionCompositionContext context);

	/**
	 * Actually reconnects the given source and target. This will only be called
	 * by the service if
	 * {@link #canReconnect(EObject, EObject, EObject, IConnectionCompositionContext)
	 * )} returned <code>true</code>. Additional information can be included in
	 * the context, which also may be null.
	 */
			boolean
			reconnect(EObject connection, S source, T target, IConnectionCompositionContext context);

	/**
	 * Returns whether the connection can be disconnected by the compositor.
	 */
	boolean canDisconnect(EObject connection);

	/**
	 * Actually disconnects the given connection. This will only be called by
	 * the service if {@link #canDisconnect(EObject)} returned <code>true</code> .
	 */
	boolean disconnect(EObject connection);
}
