/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.utils;

import static org.eclipse.emf.ecore.util.EcoreUtil.getIdentification;

import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.io.Charsets;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.BasicDiagnostic;
import org.eclipse.emf.common.util.Diagnostic;
import org.eclipse.emf.common.util.DiagnosticException;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.Diagnostician;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.osgi.framework.Bundle;

/**
 * Base class to implement serializers to load and save resources whose root element is of type
 * {@code R}.
 * 
 * @author barner
 */
public abstract class EcoreSerializerBase<R extends EObject> {

	/**
	 * {@link BasicDiagnostic} specialization that generates nice a textual representation of the
	 * validation results for a given {@link EObject}.
	 */
	private final class PrettyPrintingDiagnostics extends BasicDiagnostic {

		/** Constructs a new {@link PrettyPrintingDiagnostics}. */
		private PrettyPrintingDiagnostics(EObject obj) {
			super("EcoreSerializer", 0, "Diagnosis of " + getIdentification(obj),
					new Object[] {obj});
		}

		/** {@inheritDoc} */
		@Override
		public String getMessage() {
			return getMessage(this, "");
		}

		/**
		 * Helper for recursive implementation of {@link #getMessage()}
		 * 
		 * @param diagnostic
		 *            {@link Diagnostic} for which message should be generated.
		 * @param indent
		 *            Indentation for nice formatting of hierarchical {@link Diagnostic} structures.
		 * @return Textual representation of given {@link Diagnostic}.
		 */
		private String getMessage(Diagnostic diagnostic, String indent) {
			StringBuffer rval = new StringBuffer();

			rval.append(indent);
			if(diagnostic instanceof EcoreSerializerBase.PrettyPrintingDiagnostics) {
				rval.append(super.getMessage());
			} else {
				rval.append(diagnostic.getMessage());
			}
			rval.append("\n");

			if(diagnostic instanceof BasicDiagnostic) {
				for(Diagnostic childDiagnostic : diagnostic.getChildren()) {
					rval.append(getMessage(childDiagnostic, indent + "  "));
				}
			}

			return rval.toString();
		}
	}

	/** Namespace URI used to serialize resource. */
	private String eNsURI;

	/** {@link EPackage} that contains {@link EClass}es contained in the resource. */
	private EPackage ePackage;

	/** Factory used to build instances of elements contained in the resource. */
	private Resource.Factory resourceFactory;

	/** Encoding of resource. */
	private Charset charset;

	/** Name of schema file. */
	private String schemaFileName;

	/** Flag whether to validate model after loading and prior to saving. */
	private boolean doValidate;

	/**
	 * Flag whether to abort with an exception when validation errors are encountered when loading
	 * or saving a model. If set to {@code true}, in case of a validation error ...
	 * <ul>
	 * <li>...loading the model yields {@code null}.</li>
	 * <li>...the model is not persisted.</li>
	 * </ul>
	 * <b>Note:</b> a value of {@code false} make the validation functionality of
	 * {@link EcoreSerializerBase} <b>not thread-safe</b>.</b>
	 */
	private boolean exceptionOnValidationError;

	/**
	 * Result of last {@link #validate(EObject)} run. A value of {@code null} indicates that the
	 * model that has been saved/loaded last is valid.
	 */
	private BasicDiagnostic lastDiagnostic;

	/**
	 * Constructs a new {@link EcoreSerializerBase}.
	 * 
	 * @param eNsURI
	 *            Initializer for {@link #eNsURI}.
	 * @param ePackage
	 *            Initializer for {@link #ePackage}.
	 * @param resourceFactory
	 *            Initializer for {@link #resourceFactory}.
	 * @param charset
	 *            Initializer for {@link #charset}.
	 * @param schemaFileName
	 *            Initializer for {@link #schemaFileName}. May be {@code null}.
	 * @param doValidate
	 *            Flag whether to validate model after loading and prior to saving.
	 * @param exceptionOnValidationError
	 *            Flag whether to abort with an exception when validation errors are encountered
	 *            when loading or saving a model. <b>Note:</b> a value of {@code false} make the
	 *            validation functionality of {@link EcoreSerializerBase} <b>not
	 *            thread-safe</b>.</b>
	 */
	public EcoreSerializerBase(final String eNsURI, final EPackage ePackage,
			final Resource.Factory resourceFactory, final Charset charset,
			final String schemaFileName, final boolean doValidate,
			final boolean exceptionOnValidationError) {
		this.eNsURI = eNsURI;
		this.ePackage = ePackage;
		this.resourceFactory = resourceFactory;
		this.charset = charset;
		this.schemaFileName = schemaFileName;
		this.doValidate = doValidate;
		this.exceptionOnValidationError = exceptionOnValidationError;
	}

	/**
	 * <p>
	 * Constructs a new {@link EcoreSerializerBase} using a simplified constructor signature with
	 * the following default values:
	 * </p>
	 * <ul>
	 * <li>Charset: UTF-8</li>
	 * <li>No XSD schema file</li>
	 * <li>Model validation enabled</>
	 * <li>No exception is thrown on validation error. <b>Note:</b> In this configuration, the
	 * validation functionality is <b>not thread-safe</b>.</li>
	 * </ul>
	 * See
	 * {@link EcoreSerializerBase#EcoreSerializerBase(String, EPackage, org.eclipse.emf.ecore.resource.Resource.Factory, Charset, String, boolean, boolean)}
	 * .
	 * 
	 * @param eNsURI
	 *            Initializer for {@link #eNsURI}.
	 * @param ePackage
	 *            Initializer for {@link #ePackage}.
	 * @param resourceFactory
	 *            Initializer for {@link #resourceFactory}.
	 */
	public EcoreSerializerBase(final String eNsURI, final EPackage ePackage,
			final Resource.Factory resourceFactory) {
		this(eNsURI, ePackage, resourceFactory, Charsets.UTF_8, null, true, false);
	}

	/**
	 * Loads a resource from a file and returns its root element.
	 * 
	 * @param pathName
	 *            Path to file to be loaded
	 * @return Root element of resource.
	 * 
	 * @throws IOException
	 *             in case the model could not be loaded due to some I/O problem.
	 * @throws DiagnosticException
	 *             in case the loaded model could not be validated and the load process should be
	 *             aborted (see {@link #exceptionOnValidationError}).
	 */
	public final R load(final String pathName) throws IOException, DiagnosticException {
		return load(createLoadURI(pathName));
	}

	/**
	 * Loads a resource from a file and returns its root element.
	 * 
	 * @param uri
	 *            Path to file to be loaded
	 * @return Root element of resource.
	 * 
	 * @throws IOException
	 *             in case the model could not be loaded due to some I/O problem.
	 * @throws DiagnosticException
	 *             in case the loaded model could not be validated and the load process should be
	 *             aborted (see {@link #exceptionOnValidationError}).
	 */
	public final R load(final URI uri) throws IOException, DiagnosticException {
		ResourceSet resourceSet = createResourceSet();
		Resource resource = resourceSet.getResource(uri, true);

		if(resource.getContents().isEmpty()) {
			return null;
		}
		// get(0) is guaranteed to be safe because we just checked that the resource is not empty
		// and we pick the first element because we want the root content
		final R rootElement = getRootElement(resource.getContents().get(0));

		validate(rootElement);

		return rootElement;
	}

	/**
	 * Saves a EMF model to a file.
	 * 
	 * @param rootElement
	 *            Root element of model to serialized to XML.
	 * @param pathName
	 *            File to which configuration is to be written to.
	 * 
	 * @return {@link URI} to which EMF model has been saved.
	 * 
	 * @throws IOException
	 *             in case the model could not be saved due to some I/O problem.
	 * @throws DiagnosticException
	 *             in case the model to be saved could not be validated. Depending on the selected
	 *             policy, the model may or may not have been saved (see
	 *             {@link #exceptionOnValidationError}).
	 */
	public final URI save(final R rootElement, final String pathName) throws IOException,
			DiagnosticException {
		final URI uri = createSaveURI(pathName);
		save(rootElement, uri);

		return uri;
	}

	/**
	 * Saves a EMF model to a file.
	 * 
	 * @param rootElement
	 *            Root element of model to serialized to XML.
	 * @param uri
	 *            {@link URI} to which configuration is to be written to.
	 * 
	 * @throws IOException
	 *             in case the model could not be saved due to some I/O problem.
	 * @throws DiagnosticException
	 *             in case the model to be saved could not be validated. Depending on the selected
	 *             policy, the model may or may not have been saved (see
	 *             {@link #exceptionOnValidationError}).
	 */
	public final void save(final R rootElement, final URI uri) throws IOException,
			DiagnosticException {

		ResourceSet resourceSet = createResourceSet();
		Resource resource = resourceSet.createResource(uri);
		resource.getContents().add(rootElement);

		validate(rootElement);

		if(resource instanceof XMLResource) {
			((XMLResource)resource).setEncoding(charset.name());
		}
		resource.save(getOptions());
	}

	/**
	 * Returns the effective options to be used for serialization, i.e. the options returned by
	 * {@link #getSpecificOptions()} modified to apply some enforced options.
	 */
	private Map<Object, Object> getOptions() {
		Map<Object, Object> options = getSpecificOptions();

		// Enforce the use of extended meta-data annotations. This ensures to preserve the
		// element/attribute capitalization, the serialization type of string only elements, etc. to
		// be confirming to the original XSD.
		options.put(XMLResource.OPTION_EXTENDED_META_DATA, Boolean.TRUE);

		// If a schema file as been specifed, enforce creation of xsi:schemaLocation
		if(schemaFileName != null) {
			options.put(XMLResource.OPTION_SCHEMA_LOCATION, Boolean.TRUE);
		}

		return options;
	}

	/**
	 * Returns the options used in {@link #save(EObject, URI)}. Sub-classes may this method in
	 * order to configure the right set of options for the corresponding file format.
	 */
	protected Map<Object, Object> getSpecificOptions() {
		Map<Object, Object> options = new HashMap<Object, Object>();

		// Prefer lean files and omit default content
		options.put(XMLResource.OPTION_KEEP_DEFAULT_CONTENT, Boolean.FALSE);

		// Pre-requisite for mixed content (e.g., also for comments)
		options.put(XMLResource.OPTION_USE_LEXICAL_HANDLER, Boolean.TRUE);

		return options;
	}

	/**
	 * <p>
	 * Getter for {@link #lastDiagnostic}.
	 * </p>
	 * <p>
	 * <b>Note:</b> This method is <b>not thread-safe</b>. See {@link #exceptionOnValidationError}.
	 * </p>
	 */
	public Diagnostic getLastDiagnostic() {
		return lastDiagnostic;
	}

	/**
	 * Extracts the root element from the resources document root element. Must be implemented by
	 * sub-classes since this is specific to the corresponding file format.
	 */
	protected abstract R getRootElement(EObject contents);

	/** Converts the given {@code pathName} into a file {@link URI}. */
	protected final URI createFileURI(final String pathName) {
		return URI.createFileURI(pathName);
	}

	/**
	 * Converts the given {@code bundlePathName} into a {@link URI} within the given {@link Bundle}.
	 * 
	 * @param bundlePathName
	 *            {@link Bundle}-relative path name
	 * @param bundle
	 *            {@link Bundle} in which URI should be created
	 * @return URI to resource within the given {@link Bundle}.
	 */
	protected final URI createBundleURI(final String bundlePathName, final Bundle bundle) {

		return URI.createPlatformPluginURI(new Path(bundle.getSymbolicName())
				.append(bundlePathName).toString(), true);
	}

	/**
	 * Creates a temporary file and returns an {@link URI} pointing to it.
	 * 
	 * @param fileName
	 *            File name (=base name + extension) to be used to derive the name of the temporary
	 *            file.
	 * @param extension
	 *            File extension to be used to derive the name of the temporary file.
	 * @return {@link URI} pointing to temporary file.
	 * 
	 * @throws IOException
	 *             in case the creation of the temporary file failed.
	 */
	protected final URI createTemporaryFileURI(String fileName, String extension)
			throws IOException {

		if(!extension.startsWith(".")) {
			extension = "." + extension;
		}
		if(fileName.endsWith(extension)) {
			fileName = fileName.substring(0, fileName.length() - extension.length());
		}
		File tempFile = File.createTempFile(fileName, extension);
		tempFile.deleteOnExit();

		return URI.createFileURI(tempFile.getAbsolutePath());
	}

	/**
	 * From the given {@code pathName}, derives the URI that is actually being {@link #load(String)}
	 * ed. Sub-classes may override this method, e.g. to load files from a different context (such
	 * as a {@link Bundle}).
	 * 
	 * @throws IOException
	 */
	protected URI createLoadURI(final String pathName) throws IOException {
		return createFileURI(pathName);
	}

	/**
	 * From the given {@code pathName}, derives the URI to which the model is actually being
	 * {@link #save(EObject, String)}ed. Sub-classes may override this method, e.g. to save files
	 * from a different context (such as a {@link Bundle}).
	 * 
	 * @throws IOException
	 */
	protected URI createSaveURI(final String pathName) throws IOException {
		return createFileURI(pathName);
	}

	/**
	 * Creates a {@link ResourceSet} and configures it with the given
	 * {@link org.eclipse.emf.ecore.resource.Resource.Factory}.
	 * 
	 * @throws IOException
	 */
	private final ResourceSet createResourceSet() throws IOException {
		ResourceSet resourceSet = new ResourceSetImpl();
		resourceSet.getPackageRegistry().put(eNsURI, ePackage);

		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap()
				.put(Resource.Factory.Registry.DEFAULT_EXTENSION, resourceFactory);

		// If specified, add schema file to xsi:schemaLocation (enforces OPTION_SCHEMA_LOCATION).
		if(schemaFileName != null) {
			Resource packageResource = resourceSet.createResource(URI.createURI(schemaFileName));
			packageResource.getContents().add(ePackage);
		}

		return resourceSet;
	}

	/**
	 * In case {@link #doValidate} is {@code true}, validates the model with the given root
	 * {@link EObject}. {@link #lastDiagnostic} contains the {@link Diagnostic} indicating the
	 * validation error, or {@code null} in case the model validated successfully.
	 * 
	 * @param obj
	 *            Model to be validated.
	 * 
	 * @throws DiagnosticException
	 *             in case the model did not validate and {@link #exceptionOnValidationError} is
	 *             {@code true}.
	 */
	private void validate(final EObject obj) throws DiagnosticException {

		if(!doValidate) {
			lastDiagnostic = null;
			return;
		}

		Diagnostician diagnostician = new Diagnostician();
		Map<Object, Object> context = diagnostician.createDefaultContext();
		lastDiagnostic = new PrettyPrintingDiagnostics(obj);

		if(diagnostician.validate(obj, lastDiagnostic, context)) {
			lastDiagnostic = null;
		} else if(exceptionOnValidationError) {
			throw new DiagnosticException(lastDiagnostic);
		}
	}
}
