/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.utils;

import static org.eclipse.core.resources.ResourcesPlugin.getWorkspace;
import static org.eclipse.emf.common.util.URI.createFileURI;
import static org.eclipse.emf.common.util.URI.createURI;
import static org.eclipse.emf.ecore.util.EcoreUtil.resolveAll;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getEClassForClass;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceFactoryImpl;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.PackageNotFoundException;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceImpl;
import org.eclipse.emf.ecore.xmi.impl.XMLResourceFactoryImpl;
import org.eclipse.emf.ecore.xml.type.AnyType;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Utility class for dealing with EMF {@link Resource}s.
 * 
 * @author mou
 */
public final class ResourceUtils {

	/**
	 * {@link org.eclipse.emf.ecore.resource.Resource.Factory} that creates {@link Resource}s using
	 * the {@code id} filed of {@link IIdLabeled} model elements.
	 * 
	 * One effect of using this that {@link EReference}s are encoded based on these IDs,
	 * and hence cross-references between multiple resources are more reliable (the
	 * fragment-based encoding breaks as soon as the structure of of the referenced
	 * model changes).
	 */
	public static class KernelResourceFactory extends ResourceFactoryImpl {

		/**
		 * {@link XMIResourceImpl} used to persist models managed by the tooling kernel.
		 * 
		 * @author barner
		 */
		private final class KernelXMIResource extends XMIResourceImpl {
			/** Constructs a new {@link KernelXMIResource}. */
			private KernelXMIResource(URI uri) {
				super(uri);
			}

			/** {@inheritDoc} */
			@Override
			protected boolean useIDs() {
				return true;
			}

			/** {@inheritDoc} */
			@Override
			public String getID(EObject eObject) {
				String id = super.getID(eObject);
				if(id == null && eObject instanceof IIdLabeled) {
					id = new Integer(((IIdLabeled)eObject).getId()).toString();
					setID(eObject, id);
				}
				return id;
			}

			/** {@inheritDoc} */
			@Override
			public Map<EObject, String> getEObjectToIDMap() {
				if(eObjectToIDMap == null) {
					/*
					 * {@link IdentityHashMap} is used to implement alternative {@link EObject} ->
					 * ID maps.
					 * 
					 * This is required because {@link XMIResourceImpl} is based on {@link HashMap}s
					 * that may not work correctly for keys whose {@link #hashCode()} and
					 * {@link #equals(Object)} methods have been overridden in such a way that
					 * different instances of the same object cannot be distinguished.
					 * 
					 * Since it is not possible to provide an alternative {@link #hashCode()} and
					 * {@link #equals(Object)} implementation to a {@link HashMap}, a {@link
					 * IdentityHashMap} is used that uses {@link System#identityHashCode(Object)} to
					 * compute hash codes.
					 */
					eObjectToIDMap = new IdentityHashMap<EObject, String>();
				}

				return eObjectToIDMap;
			}

			// No override is required for getIDToEObjectMap() since its key type is String

			/** {@inheritDoc} */
			@Override
			public Map<EObject, AnyType> getEObjectToExtensionMap() {
				if(eObjectToExtensionMap == null) {
					// See getEObjectToIDMap()
					eObjectToExtensionMap = new IdentityHashMap<EObject, AnyType>();
				}
				return eObjectToExtensionMap;
			}
		}

		/** {@inheritDoc} */
		@Override
		public Resource createResource(URI uri) {
			return new KernelXMIResource(uri);
		}
	}

	/**
	 * Generate the platform URI based on the given plugin ID and an absolute
	 * path.
	 * 
	 * @param pluginId
	 *            The plugin ID
	 * @param absPath
	 *            The absolute path in the plugin
	 * @return a resource URI
	 */
	public static String getResourceURI(String pluginId, String absPath) {
		if(pluginId == null) {
			throw new NullPointerException("The plugin ID must not be null!");
		}
		if(absPath == null) {
			throw new NullPointerException("The path must not be null!");
		}
		if(!absPath.startsWith("/")) {
			throw new IllegalArgumentException("The path is not a absolute path.");
		}

		return "platform:/plugin/" + pluginId + absPath;
	}

	/**
	 * Get the string content of the resource located in the given plugin at the given path.
	 * 
	 * @param pluginId
	 *            The plugin ID
	 * @param absPath
	 *            The absolute path in the plugin
	 * @return the file content
	 */
	public static List<String> getStringContentFromResource(String pluginId, String absPath)
			throws IOException {
		URL url = new URL(getResourceURI(pluginId, absPath));
		List<String> result = new ArrayList<String>();
		BufferedReader r = new BufferedReader(new InputStreamReader(url.openStream()));
		String l;
		while((l = r.readLine()) != null) {
			result.add(l);
		}
		return result;
	}

	/**
	 * <p>
	 * Determines the {@link ITopLevelElement} which shares the same {@link ResourceSet} as the a
	 * {@link Resource} with the specified {@link URI}. If no such {@link Resource} has been loaded,
	 * {@code null} is returned.
	 * </p>
	 * <p>
	 * This is useful to determine to which model managed by the Kernel a given (external) model is
	 * linked.
	 * </p>
	 * 
	 * @param uri
	 *            URI for which the corresponding {@link ITopLevelElement} should be determined.
	 * 
	 * @return {@link ITopLevelElement} which shares the same {@link ResourceSet} as the a
	 *         {@link Resource} with the specified {@link URI} ({@code null} if there is none).
	 */
	public static ITopLevelElement getTopLevelElement(URI uri) {
		List<ITopLevelElement> elems = IPersistencyService.getInstance().getTopLevelElements();

		for(ITopLevelElement topLevelElement : elems) {
			for(Resource resource : topLevelElement.getResourceSet().getResources()) {
				if(resource.getURI().equals(uri)) {
					return topLevelElement;
				}
			}
		}

		return null;
	}

	/**
	 * Obtains a model of a given type from a {@code resourceSet}, or {@code null} if it could not
	 * be found.
	 * 
	 * @param resourceSet
	 *            {@link ResourceSet} from which model of given type should be determined
	 * @param clazz
	 *            Type of model to be returned.
	 * @return Model of a given type from a {@code resourceSet}, or {@code null} if it could not
	 *         be found.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends EObject> T getModel(ResourceSet resourceSet, Class<T> clazz) {
		resolveAll(resourceSet);
		for(Resource resource : resourceSet.getResources()) {
			for(EObject obj : resource.getContents()) {
				if(clazz.isAssignableFrom(obj.getClass())) {
					return (T)obj;
				}
			}
		}

		return null;
	}

	/**
	 * Obtains a model of a given type from a {@code resourceSet}, and creates the model if it does
	 * not exist yet (and adds it to the given {@code resourceSet}). Returns {@code null} if the
	 * model could not be created.
	 * 
	 * @param resourceSet
	 *            {@link ResourceSet} from which model of given type should be determined
	 * @param clazz
	 *            Type of model to be returned.
	 * @param factory
	 *            {@link EFactory} to create root model element in case the model does not exist in
	 *            the {@code resourceSet} yet.
	 * @param fileExtension
	 *            File extension to be used in case a new resource has to be created and added to
	 *            the given given {@code resourceSet}. In case this parameter is {@code null}, this
	 *            method effectively behaves like {@link #getModel(ResourceSet, Class)}.
	 * @param referenceClazz
	 *            Type of root model element whose {@link Resource} {@link URI} should be used to
	 *            derive the base name in case a new {@link Resource} has to be created. In case
	 *            this parameter is {@code null}, the {@link URI} of the first {@link Resource} in
	 *            the given {@link ResourceSet} is used.
	 * 
	 * @return Model of a given type, or {@code null} if it could not be created.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends EObject, R extends EObject> T getModel(ResourceSet resourceSet,
			Class<T> clazz, EFactory factory, String fileExtension, Class<R> referenceClazz) {
		T existingModel = getModel(resourceSet, clazz);

		if(existingModel != null) {
			return existingModel;
		}

		if(fileExtension != null) {
			resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap()
					.put(fileExtension, new XMLResourceFactoryImpl());
		}

		URI uri = null;
		EObject obj = null;
		if(referenceClazz != null) {
			obj = getModel(resourceSet, referenceClazz);
			if(obj != null) {
				uri = obj.eResource().getURI();
			}
		} else if(!resourceSet.getResources().isEmpty()) {
			// "resourceSet" is guaranteed to be non-empty
			Resource resource = resourceSet.getResources().get(0);
			uri = resource.getURI();

			// Resources that are in the resource set are expected to be non-empty
			obj = resource.getContents().get(0);
		}

		if(factory != null && fileExtension != null && uri != null && obj != null) {
			uri = uri.trimFileExtension().appendFileExtension(fileExtension);
			Resource resource = resourceSet.createResource(uri);
			final T newModel = (T)factory.create(getEClassForClass(clazz));

			ITopLevelElement modelContext =
					IPersistencyService.getInstance().getTopLevelElementFor(obj);
			modelContext.runAsCommand(new Runnable() {
				@Override
				public void run() {
					resource.getContents().add(newModel);
				}
			});

			return newModel;
		}

		return null;
	}

	/**
	 * Determines the {@link IFile} in the Eclipse workspace that corresponds to the given
	 * {@link Resource}.
	 * 
	 * @param resource
	 *            Resource for which {@link IFile} should be determined.
	 * @return {@link IFile} in the Eclipse workspace that corresponds to {@code resource}.
	 */
	public static IFile getIFile(Resource resource) {
		final URI uri = resource.getURI();
		String resourcePath = uri.toPlatformString(true);
		final IWorkspaceRoot workspaceRoot = getWorkspace().getRoot();

		final IResource member = workspaceRoot.findMember(resourcePath);

		// EMF resources always correspond to IFile IResources
		return (IFile)member;
	}

	/**
	 * For the given {@link Resource}, returns the absolute (file) {@link URI}.
	 * 
	 * @param resource
	 *            {@link Resource} for which absolute (file) {@link URI} should be determined.
	 * @return Absolute (file) {@link URI} of given {@link Resource}.
	 */
	public static URI getAbsoluteResourceURI(Resource resource) {
		IFile file = getIFile(resource);
		final IWorkspaceRoot workspaceRoot = getWorkspace().getRoot();
		String absoluteResourcePath =
				workspaceRoot.getLocation().append(file.getFullPath()).toFile().toURI().toString();
		final URI resourceUri = createURI(absoluteResourcePath, true);
		return resourceUri;
	}

	/**
	 * Determines {@link URI}s of the resources that directly or indirectly the given
	 * {@link Resource}. It investigates all files that reside in the same directory as
	 * {@code resource}.
	 * 
	 * @param resource
	 *            {@link Resource} for which {@link URI}s of directly or indirectly referencing
	 *            resources should be determined.
	 * @param includeSelf
	 *            Flag if the {@link URI} of the given {@code resource} should be contained in the
	 *            result.
	 * @param progressMonitor
	 *            {@link IProgressMonitor} (may be {@code null}). Loading all resources in a
	 *            directory may be a long-running operation.
	 * @return {@link URI}s of directly or indirectly referencing resources, possibly including the
	 *         {@link URI} of the given {@code resource} itself (see {@code includeSelf}).
	 */
	public static Collection<URI> getReferencingResourceURIs(Resource resource,
			boolean includeSelf, IProgressMonitor progressMonitor) {
		IFile file = getIFile(resource);

		final URI resourceUri = getAbsoluteResourceURI(resource);

		Set<URI> uris = new HashSet<URI>();
		uris.add(resourceUri);

		IContainer parent = file.getParent();
		try {
			final IResource[] members = parent.members();
			if(progressMonitor != null) {
				progressMonitor.beginTask("Determining referencing resources.", members.length);
			}

			for(IResource member : members) {
				if(member.equals(file) || member instanceof IFolder) {
					continue;
				}

				ResourceSet resourceSet = new ResourceSetImpl();
				try {
					URI siblingUri = createFileURI(member.getLocation().toString());
					Resource siblingResource = resourceSet.getResource(siblingUri, true);
					resolveAll(resourceSet);

					for(Resource currentResource : resourceSet.getResources()) {
						final URI currentUri = currentResource.getURI();
						if(uris.contains(currentUri)) {
							uris.addAll(resourceSet.getResources().stream().map(r -> r.getURI())
									.collect(Collectors.toSet()));
							break;
						}
					}

					siblingResource.unload();
				} catch(Exception e) {
					if(!(e.getCause() instanceof PackageNotFoundException)) {
						error(ToolingKernelActivator.getDefault(),
								"Failed to load \"" + member.getName(), e);
					}
				}

				if(progressMonitor != null) {
					progressMonitor.worked(1);
				}
			}
		} catch(CoreException e) {
			error(ToolingKernelActivator.getDefault(),
					"Failed to determine members of \"" + parent.getName(), e);
		}

		if(!includeSelf) {
			uris.remove(resourceUri);
		}

		return uris;
	}
}
